#============================================================= -*-perl-*-
#
# Template::Base
#
# DESCRIPTION
#   Base class for other Template::* modules
#
# AUTHOR
#   Andy Wardley   <abw@kfs.org>
#
# COPYRIGHT
#   Copyright (C) 2001 Andy Wardley.  All Rights Reserved.
#
#   This module is free software; you can redistribute it and/or
#   modify it under the same terms as Perl itself.
#
# REVISION
#   $Id: Base.pm,v 1.2 2001/06/19 16:15:25 abw Exp $
#
#========================================================================

package Template::Base;

use strict;
use vars qw( $VERSION $DEBUG $ERROR $ECLASS $ETYPE $ERRSTR $SNIPPET );
use Template::Exception;

$VERSION = sprintf("%d.%02d", q$Revision: 1.2 $ =~ /(\d+)\.(\d+)/);
$ECLASS  = 'Template::Exception';
$ETYPE   = 'undef';
$SNIPPET = 32;
$ERRSTR  = {
    no_factory => 'no factory defined',
};


#------------------------------------------------------------------------
# new(@config)
# new(\%config)
#
# General purpose constructor method which expects a hash reference of 
# configuration parameters, or a list of name => value pairs which are 
# folded into a hash.  Blesses a hash into an object and calls its 
# init() method, passing the parameter hash reference.  Returns a new
# object derived from Template::Base, or undef on error.
#
# TODO: should look for @ARGS in derived class package and shift those 
# items off the stack into member variables before converting rest of 
# arguments to config hash.
#------------------------------------------------------------------------

sub new {
    my $class  = shift;
    my ($argnames, @args, $arg, $cfg);
#    $class->error('');		# always clear package $ERROR var?

    {	no strict qw( refs );
	$argnames = \@{"$class\::ARGS"} || [ ];
    }

    # shift off all mandatory args, returning error if undefined or null
    foreach $arg (@$argnames) {
	return $class->error("no $arg specified")
	    unless ($cfg = shift);
	push(@args, $cfg);
    }

    # fold all remaining args into a hash, or use provided hash ref
    $cfg  = defined $_[0] && UNIVERSAL::isa($_[0], 'HASH') ? shift : { @_ };

    my $self = bless {
	map { ($_ => shift @args) } @$argnames,
	_ERROR  => ''
    }, $class;

    return $self->init($cfg)
	|| $class->error($self->error());
}


#------------------------------------------------------------------------
# init()
#
# Initialisation method called by the new() constructor and passing a 
# reference to a hash array containing any configuration items specified
# as constructor arguments.  Should return $self on success or undef on 
# error, via a call to the error() method to set the error message.
#------------------------------------------------------------------------

sub init {
    my ($self, $config) = @_;
    return $self;
}

#------------------------------------------------------------------------
# factory($new_factory)
#
# Can be called as a class method to retrieve or update the $FACTORY
# package variable or as an object method to do likewise with the 
# $self->{ _FACTORY } item.  Returns the current value when called
# without arguments.  A new factory class name or object reference
# can be passed as an argument to update the relevant variable.
#------------------------------------------------------------------------

sub factory {
    my $self = shift;
    my $factory;

    { 
	no strict qw( refs );
	if (ref $self && $self->{ _FACTORY }) {
	    $factory = \$self->{ _FACTORY };
	}
	else {
	    $factory = \$Template::FACTORY;
	}
    }
    if (@_) {
	return ($$factory = shift);
    }
    else {
	return $$factory
	    || $self->errstr('no_factory');
    }
}


#------------------------------------------------------------------------
# create($module_type)
#
# Delegate to factory create() method.
#------------------------------------------------------------------------

sub create {
    my $self = shift;

    my $factory = $self->factory()
	|| return $self->errstr( no_factory => ' to create() with' );

    return $factory->create(@_)
	|| $self->error($factory->error());
}

#------------------------------------------------------------------------
# mandatory(\%config, \@fields)
# mandatory(\%config, $field1, $field2, ...)
#
# Copies mandatory fields, specified in $fields list, from $config 
# hash ref to $self.  Returns reference to $self on success or raises 
# error and returns undef if any field is undefined.
#------------------------------------------------------------------------

sub mandatory {
    my $self   = shift;
    my $config = shift;
    my $fields = ref $_[0] eq 'ARRAY' ? shift : [ @_ ];

    foreach my $f (@$fields) {
	return $self->error("$f attribute not specified")
	    unless defined $config->{ $f };
	$self->{ $f } = $config->{ $f };
    }
    return $self;
}


#------------------------------------------------------------------------
# optional(\%config, \@fields)
# optional(\%config, $field1, $field2)
# optional(\%config, \%fields)
#
# Copies optional fields, specified in $fields, from $config hash ref 
# to $self.  If $fields is a hash reference then the corresponding value
# for any field will be used as a default if otherwise undefined in 
# $config.  Returns reference to $self.
#------------------------------------------------------------------------

sub optional {
    my $self   = shift;
    my $config = shift;
    my $fields = ref $_[0] ? shift : [ @_ ];
    my ($key, $val);
    
    $fields = { map { ($_ => undef) } @$fields } 
	if ref $fields eq 'ARRAY';

    while (($key, $val) = each %$fields) {
	$self->{ $key } = exists $config->{ $key }
	    ? $config->{ $key } : $val;
    }

    return $self;
}


#------------------------------------------------------------------------
# error()
# error($msg, ...)
# 
# May be called as a class or object method to set or retrieve the 
# package variable $ERROR (class method) or internal member 
# $self->{ _ERROR } (object method).  The presence of parameters indicates
# that the error value should be set.  Undef is then returned.  In the
# abscence of parameters, the current error value is returned.
#------------------------------------------------------------------------

sub error {
    my $self = shift;
    my $errvar;

    { 
	no strict qw( refs );
	$errvar = ref $self ? \$self->{ _ERROR } : \${"$self\::ERROR"};
    }
    if (@_) {
	# don't join if first arg is an object (may force stringification)
	$$errvar = ref($_[0]) ? shift : join('', @_);
	return undef;
    }
    else {
	return $$errvar;
    }
}

#------------------------------------------------------------------------
# errstr($id, $msg)
# 
# This can be called as a class or object method to raise an error via
# error() using a pre-defined error message in the hash array referenced
# by the $ERRSTR package variable in the derived class.  The first 
# argument should specify a key into the hash (otherwise 'default' is
# tried).  Further optional arguments may also be provided to specify
# additional error messages.
#------------------------------------------------------------------------

sub errstr {
    my $self  = shift;
    my $class = ref $self || $self;
    my $id    = shift;
    my $msg   = join('', @_);
    my ($errstr, $error);

    { 
	no strict qw( refs );
	$errstr = ${"$class\::ERRSTR"};
    }

    $error = $errstr->{ $id }
	  || $errstr->{ default }
	  || $ERRSTR->{ $id }
	  || "unexpected error [$id] from $class";
    
    $error .= $msg if length $msg;
    $self->error($error);

}


#------------------------------------------------------------------------
# throw($type, $info, \$output)          [% THROW errtype "Error info" %]
#
# Throws a Template::Exception object by calling die().  This method
# may be passed a reference to an existing Template::Exception object;
# a single value containing an error message which is used to
# instantiate a Template::Exception of type 'undef'; or a pair of
# values representing the exception type and info from which a
# Template::Exception object is instantiated.  e.g.
#
#   $context->throw($exception);
#   $context->throw("I'm sorry Dave, I can't do that");
#   $context->throw('denied', "I'm sorry Dave, I can't do that");
#
# An optional third parameter can be supplied in the last case which 
# is a reference to the current output buffer containing the results
# of processing the template up to the point at which the exception 
# was thrown.  The RETURN and STOP directives, for example, use this 
# to propagate output back to the user, but it can safely be ignored
# in most cases.
# 
# This method rides on a one-way ticket to die() oblivion.  It does not 
# return in any real sense of the word, but should get caught by a 
# surrounding eval { } block (e.g. a BLOCK or TRY) and handled 
# accordingly, or returned to the caller as an uncaught exception.
#------------------------------------------------------------------------

sub throw {
    my ($self, $error, $info, $output) = @_;
    local $" = ', ';

    require Template::Exception;

    # die! die! die!
    if (UNIVERSAL::isa($error, $ECLASS)) {
	$self->DEBUG("throwing existing exception [@$error]\n");
	die $error;
    }
    elsif (defined $info) {
	$self->DEBUG("throwing new exception [$error] [$info]\n");
	die $ECLASS->new($error, $info, $output);
    }
    else {
	no strict 'refs';
	$error ||= '';
	# look for $ETYPE in derived class
	my $class = ref $self;
	my $ename = ${"$class\::ETYPE"} || $ETYPE;
	$self->DEBUG("throwing an $ename exception [$error]\n");
	die $ECLASS->new($ename, $error, $output);
    }

    # not reached
}


#========================================================================
# debug methods for derived classes
#========================================================================

#------------------------------------------------------------------------
# DEBUG(@msg)
#
# Raise debug message.
#------------------------------------------------------------------------

sub DEBUG {
    my $self = shift;
    print STDERR "DEBUG: ", join('', @_);
}


#------------------------------------------------------------------------
# ID
#
# Return a useful string describing $self for debugging purposes.
#------------------------------------------------------------------------

sub ID {
    my $self = shift;
    return $self;
}


#------------------------------------------------------------------------
# INSPECT
#
# Return a string detailing internals and/or current object status.
#------------------------------------------------------------------------

sub INSPECT {
    my $self = shift;
    my $output = $self->ID . " => {\n";
    while (my ($key, $value) = each %$self) {
	$value = '<undef>' unless defined $value;
	$output .= sprintf("  %-12s => %s\n", $key, $value);
    }
    $output .= '}';
    return $output;
}


#========================================================================
# general purpose debug/utility methods
#======================================================================== 

#------------------------------------------------------------------------
# _inspect_ref($ref)
#
# Tries to Do the Right Thing to inspect a reference.
#------------------------------------------------------------------------

sub _inspect {
    my $self = shift;
    $self->_inspect_ref(@_);
}

sub _inspect_ref {
    my ($self, $ref) = @_;
    if (UNIVERSAL::isa($ref, 'HASH')) {
	return $self->_inspect_hash($ref);
    }
    elsif (UNIVERSAL::isa($ref, 'LIST')) {
	return $self->_inspect_list($ref);
    }
    elsif (UNIVERSAL::can($ref, 'INSPECT')) {
	return $ref->INSPECT();
    }
    else {
	return $ref;
    }
}

#------------------------------------------------------------------------
# _inspect_hash(\%hash)
#
# Returns a string representing the contents of the hash.
#------------------------------------------------------------------------

sub _inspect_hash {
    my $self = shift;
    my $hash = shift;
    return '{ ' . join(', ', map { 
	my $val = $hash->{ $_};
	$val = '<undef>' unless defined $val;
	$val = $self->_inspect_ref($val) if ref $val;
	"$_ => $val" 
    } keys %$hash) . ' }';
}

#------------------------------------------------------------------------
# _inspect_list(\@list)
#
# Returns a string representing the contents of the list.
#------------------------------------------------------------------------

sub _inspect_list {
    my ($self, $list) = @_;
    return '[ ' . join(', ', map { 
	my $val = $_;
	$val = '<undef>' unless defined $val;
	$val = $self->_inspect_ref($val) if ref $val;
	$val;
    } @$list) . ' ]';
}

#------------------------------------------------------------------------
# _inspect_text($text, $length)
#
# Truncates the text to $length (or $SNIPPET) length and appends '...' 
#------------------------------------------------------------------------

sub _inspect_text {
    my ($self, $text, $length) = @_;
    $length ||= $SNIPPET;
    my $snippet = substr($text, 0, $length);
    $snippet .= '...' if length $text > $length;
    $snippet =~ s/\n/\\n/g;
    return $snippet;
}

1;

