#============================================================= -*-Perl-*-
#
# Template::Compiler
#
# DESCRIPTION
#   Compiler module implementing methods for generating the results
#   of satisfied production rules within the parser.  This default
#   and base class compiler module generates Perl code.
# 
# AUTHOR
#   Andy Wardley <abw@kfs.org>
#
# COPYRIGHT
#   Copyright (C) 1996-2001 Andy Wardley.  All Rights Reserved.
#   Copyright (C) 1998-2001 Canon Research Centre Europe Ltd.
#
#   This module is free software; you can redistribute it and/or
#   modify it under the same terms as Perl itself.
#
# REVISION
#   $Id
#
#========================================================================

package Template::Compiler;

require 5.005;

use strict;
use vars qw( $VERSION $DEBUG $ERROR $ETYPE $DIRECTIVES );
use base qw( Template::Base );

$VERSION = sprintf("%d.%02d", q$Revision: 1.1 $ =~ /(\d+)\.(\d+)/);
$DEBUG   = 0 unless defined $DEBUG;
$ETYPE   = 'compiler';
$ERROR   = '';
$DIRECTIVES = {
    INCLUDE  => \&include,
    GET	     => \&get,
    CONSTANT => \&constant,
};

sub template_handler {
    my $self = shift;

    # build list of callbacks, e.g. INCLUDE => [ $self, \&include ]
    # for handler to make when directives are encountered
    my $dirs = { 
	map { ($_, [ $self, $DIRECTIVES->{ $_ } ]) }
	keys %$DIRECTIVES,
    };
    my $stash = $self->create('stash');
    $self->{ _CONSTANTS } = $stash;

    return $self->create('parser_handler', DIRECTIVES => $dirs);
}

#------------------------------------------------------------------------

sub number {
    # my ($self, $number) = @_;
    return $_[1];
}

sub literal {
    my ($self, $literal) = @_;
    return "'$literal'";
}

sub quoted {
    my ($self, $items) = @_;
    return '( "" . ' . join(' . ', @$items) . ')';
}

sub variable {
    my ($self, $term) = @_;
    my $stash = $self->{ _CONSTANTS };
    my $lookup;

    $self->DEBUG('var(', $self->_inspect($term), ")\n")
	if $DEBUG;

    return "''" unless @$term;

    if (scalar @$term <= 2 && ! $term->[1]) {
	$term = $term->[0];
	$lookup = $term;
	for ($lookup) {
	    s/^'//;
	    s/'$//;
	}
	# look for constants
	return "'" . $stash->get($lookup) . "'"
	    if exists $stash->{ $lookup };
    }
    else {
	$lookup = $term->[0];
	$term = '[' . join(', ', @$term) . ']';
	for ($lookup) {
	    s/^'//;
	    s/'$//;
	}
	return "'" . (eval "\$stash->get($term)") . "'"
	    if exists $stash->{ $lookup };
    }
    return "\$stash->get($term)";
}

sub assign {
    my ($self, $var, $value) = @_;
    $self->DEBUG('assign(', $self->_inspect($var), ', ', 
		 $self->_inspect($value), ")\n")
	if $DEBUG;

    # TODO: look for constants

    if (ref $var eq 'ARRAY') {
	if (scalar @$var == 2 && ! $var->[1]) {
	    $var = $var->[0];
	}
	else {
	    $var = '[' . join(', ', @$var) . ']';
	}
    }
#    $val .= ', 1' if $default;
    return "\$stash->set($var, $value)";
}

sub setlist {
    my ($self, $setlist) = @_;

    $self->DEBUG('setlist(', $self->_inspect($setlist), ")\n")
	if $DEBUG;

    my $output = '';

    foreach (@$setlist) {
	my ($var, $value) = @$_;
	$output .= $self->assign($var, $value) . ";\n";
    }
#    return '$stash->update([ ' . join(', ', @output) . '])';
    chomp $output;
    return $output;
}

sub qwlist {
    my ($self, $term) = @_;
    $self->DEBUG('qwlist(', $self->_inspect($term), ")\n")
	if $DEBUG;
    return "[ qw( $term ) ]";
}

sub list {
    my ($self, $term) = @_;
    $self->DEBUG('list(', $self->_inspect($term), ")\n")
	if $DEBUG;
    return '[ '
	. join(', ', map { ref eq 'ARRAY' ?  "$_->[0] .. $_->[1]" : $_ } @$term)
	. ' ]';
}

sub hash {
    my ($self, $term) = @_;
    $self->DEBUG('hash(', $self->_inspect($term), ")\n")
	if $DEBUG;
    return "'HASH: TODO'";
}

sub args {
    my ($self, $args) = @_;
    $self->DEBUG('args(', $self->_inspect($args), ")\n")
	if $DEBUG;
    return "'ARGS: TODO'";
}

#------------------------------------------------------------------------

sub block {
    my ($self, $content) = @_;
    local $" = ', ';
    return "do { @$content }";
}


#------------------------------------------------------------------------

sub constant {
    my ($self, $parser, $textref) = @_;
    my $stash = $self->{ _CONSTANTS };

    $$textref =~ s/^\s*constant\s+//gi;
    
    $self->DEBUG("constant($$textref)\n")
	if $DEBUG;

    my $vars = $parser->parse_hash($textref);
#    my $perl = '$stash->update([ ' . join(', ', @$vars) . '])';

#    my $perl = 'print "stash: $stash\n";';
    while (@$vars) {
	my ($var, $val) = splice(@$vars, 0, 2);
	$self->DEBUG("eval: [$var] [$val]\n");
	$var = eval $var;
	$val = eval $val;
	$stash->set($var, $val);
	$self->DEBUG("CONSTANT: set($var => $val)\n");
	if ($var eq 'args') {
	    my $args = $stash->get('args');
	    $self->DEBUG("ARGS: [@$args]\n");
	}
    }
    return '';
}


sub include {
    my ($self, $parser, $directive) = @_;
    $self->DEBUG("include($directive)\n")
	if $DEBUG;
    return "\$view->include($directive)";
}

sub get {
    my ($self, $parser, $directive) = @_;
    $self->DEBUG("get($directive)\n")
	if $DEBUG;
    return "\$stash->get($directive)";
}

#========================================================================
# debug/housekeeping methods
#========================================================================

sub ID {
    my $self = shift;
    my $ref = "$self";
    $ref = $1
	if $ref =~ /\((.*?)\)/;
    return "parser_backend<$ref>";
}


1;

__END__

