#============================================================= -*-perl-*-
#
# Template::Context
#
# DESCRIPTION
#   Object class to represent a runtime template processing context.
#
# AUTHOR
#   Andy Wardley   <abw@kfs.org>
#
# COPYRIGHT
#   Copyright (C) 2001 Andy Wardley.  All Rights Reserved.
#
#   This module is free software; you can redistribute it and/or
#   modify it under the same terms as Perl itself.
#
# REVISION
#   $Id: Context.pm,v 1.2 2001/06/19 16:15:25 abw Exp $
#
#========================================================================

package Template::Context;

require 5.005;

use strict;
use Template::Base;
use vars qw( $VERSION $DEBUG $ETYPE );
use base qw( Template::Base );

$VERSION = sprintf("%d.%02d", q$Revision: 1.2 $ =~ /(\d+)\.(\d+)/);
$DEBUG   = 0 unless defined $DEBUG;
$ETYPE   = 'context';


#------------------------------------------------------------------------
# engage($service, @args)
# 
# Engage the context to a service.
#------------------------------------------------------------------------

sub engage {
    my ($self, $service, @args) = @_;
    $self->DEBUG("$self engaging to $service\n") if $DEBUG;
    $self->notify( engage => @args ) || return undef
	if $self->{ _NOTIFY_ENGAGE };
    $self->{ _SERVICE } = $service;
    return 1;
}


#------------------------------------------------------------------------
# disengage($service, @args)
# 
# Disengage the context from a service.
#------------------------------------------------------------------------

sub disengage {
    my ($self, $service, @args) = @_;
    $self->DEBUG("$self disengaging from $service\n") if $DEBUG;
    $self->notify( disengage => @args ) || return undef
	if $self->{ _NOTIFY_DISENGAGE };
    undef $self->{ _SERVICE };
    return 1;
}


#------------------------------------------------------------------------
# switch($new_context, @args)
#
# Switch contexts from the current one to the new context specified
# by name as the first argument.  Calls the service to which the 
# current context is engaged requesting a context switch.  This then
# makes a callback to the current context disengage() method and
# the new context engage() method.
#
# NOTE: this method is called at template runtime so it throws errors
# rather than returning undef (for now at least, but subject to change).
#------------------------------------------------------------------------

sub switch {
    my ($self, $context, @args) = @_;
    my $service = $self->{ _SERVICE } 
	|| return $self->throw("no service");

    $self->DEBUG("switching context to $context\n") if $DEBUG;

    my $new = $service->context($context, @args)
	|| return $self->error($service->error());

    $self->DEBUG("switching to new context: $new\n") if $DEBUG;

    return $new;
}


#------------------------------------------------------------------------
# acquire($name, $facility, @args)
#
# Have the context acquire a facility by calling the acquire() method
# of the $facility object passing a reference to $self and any additional
# arguments passed.  If the method returns true then the facility 
# reference is stored in the internal _FACILITY hash indexed by $name.
#------------------------------------------------------------------------

sub acquire {
    my ($self, $name, $facility, @args) = @_;
    my $facilities = $self->{ _FACILITY } ||= { };
    
    $self->DEBUG("$self acquiring $name => $facility\n") if $DEBUG;

    $facility->acquire($self, @args)
	|| return $self->error($facility->error());

    $facilities->{ $name } = $facility;

    return 1;
}


#------------------------------------------------------------------------
# release($name)
#
# Release the facility indexed by $name.
#------------------------------------------------------------------------

sub release {
    my ($self, $name) = @_;
    my $facility = $self->{ _FACILITY }->{ $name }
	|| return $self->no_facility_error($name);

    $self->DEBUG("$self releasing $name => $facility\n") if $DEBUG;

    $facility->release($self)
	|| return $self->error($facility->error());

    undef $self->{ _FACILITY }->{ $name };

    return 1;
}


#------------------------------------------------------------------------
# release_all()
#
# Release all facilities currently acquired by this context.
#------------------------------------------------------------------------

sub release_all {
    my $self = shift;
    my $facilities = $self->{ _FACILITY } || return 1;
    my @errors;

    $self->DEBUG("$self releasing all facilities\n") if $DEBUG;

    while (my($name, $facility) = each %$facilities) {
	$facility->release($self)
	    || push(@errors, $facility->error());
	undef $facilities->{ $name };
    }

    return @errors ? $self->error(@errors) : 1;
}


#------------------------------------------------------------------------
# facility($name)
# 
# Return the facility index by $name.  Returns undef if the facility
# isn't defined and sets an appropriate error message.
#------------------------------------------------------------------------

sub facility {
    my ($self, $name) = @_;
    return $self->{ _FACILITY }->{ $name }
	|| $self->no_facility_error($name);
}


#------------------------------------------------------------------------
# facilities($name1, $name2, ...)
# 
# Return one or more facilites specfied by name as per facility().  When
# called in list context, the method returns a list of the facilites
# specified.  When called in scalar context, returns a reference to the
# same list.  Returns undef if any of the facilities are not defined 
# and sets an appropriate error message.  If no facility names are 
# specified then the method returns a reference to the internal
# _FACILITES hash.  If you access and subsequently update the entries
# in this hash then you do do at your own peril!
#------------------------------------------------------------------------

sub facilities {
    my $self = shift;
    my $facilities = $self->{ _FACILITY } ||= { };
    return $facilities unless @_;

    my ($name, $facility, @result);
    while (@_) {
	$name = shift;
	return $self->no_facility_error($name)
	    unless ($facility = $facilities->{ $name });
	push(@result, $facility);
    }

    return wantarray ? @result : \@result;
}


#------------------------------------------------------------------------
# observe($event, $object, $method, @args)
#
# Add an observer in the form of an object reference, the name of a 
# method to call and any arguments to pass, to the notification
# list denoted by $event.
#------------------------------------------------------------------------

sub observe {
    my ($self, $event, $object, $method, @args) = @_;
    my $qname = "_NOTIFY_\U$event";
    my $queue = $self->{ $qname } ||= [ ];
    push(@$queue, [ $object, $method, @args ]);
    return 1;
}

#------------------------------------------------------------------------
# notify($event, @args)
#
# Notify all observers registered to receive notification of a particular
# event 
#------------------------------------------------------------------------

sub notify {
    my ($self, $event, @args) = @_;
    my $qname = "_NOTIFY_\U$event";
    my $queue = $self->{ $qname } 
	|| return $self->error("$qname: no such notification event");
    foreach $event (@$queue) {
	my ($object, $method, @objargs) = @$event;
	$object->$method(@objargs, @args)
	    || return $self->error($object->error());
    }
    return 1;
}


#------------------------------------------------------------------------
# no_facility_error($name)
#
# Method for reporting errors indicating a requested facility is not
# defined.
#------------------------------------------------------------------------

sub no_facility_error {
    my ($self, $name) = @_;
    $name = '<undef>' unless defined $name;
    return $self->error("$name: no such facility");
}


sub DESTROY {
    my $self = shift;
    my $facilities = $self->{ _FACILITY };
    $self->release_all()
	if $facilities && %$facilities;
}

    
1;

__END__

throw()
catch()
localise()
delocalise()
define_block()
define_filter()
reset()
stash()
