#============================================================= -*-perl-*-
#
# Template::Factory
#
# DESCRIPTION
#   Factory module for instantiating other Template::* modules
#
# AUTHOR
#   Andy Wardley   <abw@kfs.org>
#
# COPYRIGHT
#   Copyright (C) 2001 Andy Wardley.  All Rights Reserved.
#
#   This module is free software; you can redistribute it and/or
#   modify it under the same terms as Perl itself.
#
# REVISION
#   $Id: Factory.pm,v 1.1 2001/06/19 16:15:25 abw Exp $
#
#========================================================================

package Template::Factory;

use strict;
use Template::Base;
use base qw( Template::Base );
use vars qw( $VERSION $AUTOLOAD $DEBUG $ERROR $ETYPE $MODULES );

$VERSION = sprintf("%d.%02d", q$Revision: 1.1 $ =~ /(\d+)\.(\d+)/);
$DEBUG   = 0 unless defined $DEBUG;
$ERROR   = '';
$ETYPE   = 'factory';
$MODULES  = {
    exception      => 'Template::Exception',
    parser_handler => 'Template::Parser::Handler',
    compiler       => 'Template::Compiler',
    contract       => 'Template::Contract',
    stash          => 'Template::Stash::v3',
};


sub init {
    my ($self, $config) = @_;
    my $class = ref $self;
    bless { %$MODULES, %$config }, $class;
}

#------------------------------------------------------------------------
# create($module_type)
#
# Look up module name for a given type in $MODULES hash if called as a 
# class method, or $self hash if called as an object method.  Load the 
# module via load() method and then instantiate an object via new()
#------------------------------------------------------------------------

sub create {
    my $self = shift;
    my $type = shift;

    $self->DEBUG($self->ID, "->create('$type')\n")
	if $DEBUG;

    my $module;
    if (ref $self) {
	$module = $self->{ $type };
    }
    else {
	$module = $MODULES->{ $type };
    }
    return $self->error("module not recognised: '$type'")
	unless $module;

    return undef unless $self->load($module);

    return $module->new(@_) 
	|| $self->error("failed to create $type: ", $module->error);
}


#------------------------------------------------------------------------
# adopt($module_type, $object, \@args)
#
# Look up module name for a given type in $MODULES hash if called as a 
# class method, or $self hash if called as an object method.  Load the 
# module via load() method, rebless $object into the new module class
# then call its new init() method passing a reference to an optional 
# hash reference of configuration options.
#------------------------------------------------------------------------

sub adopt {
    my ($self, $type, $object, $config) = @_;

    $self->DEBUG($self->ID, "->adopt('$type', $object)\n")
	if $DEBUG;

    my $module;
    if (ref $self) {
	$module = $self->{ $type };
    }
    else {
	$module = $MODULES->{ $type };
    }
    return $self->error("module not recognised: '$type'")
	unless $module;

    return undef unless $self->load($module);
    bless $object, $module;

    return $object->init($config)
	|| $self->error($object->error());
}


#------------------------------------------------------------------------
# load($module)
#
# Load a module via require().  Any occurences of '::' in the module name
# are be converted to '/' and '.pm' is appended.  Returns 1 on success
# or undef on error.  Use $class->error() to examine the error string.
#------------------------------------------------------------------------

sub load {
    my ($self, $module) = @_;
    $module =~ s[::][/]g;
    $module .= '.pm';
    $self->DEBUG($self->ID, "->require('$module')\n")
	if $DEBUG;
    eval {
	require $module;
    };
    return $@ ? $self->error("failed to load $module: $@") : 1;
}


#------------------------------------------------------------------------
# AUTOLOAD
#
# Map method calls of the form $modules->parser(....) to
# $modules->create('parser', ...)
#------------------------------------------------------------------------

sub AUTOLOAD {
    my $self = shift;
    my $item = $AUTOLOAD;
    $item =~ s/.*:://;
    return if $item eq 'DESTROY';
    $self->create($item, @_);
}
	
1;

