#============================================================= -*-Perl-*-
#
# Template::Parser
#
# DESCRIPTION
#   Parser module for the Template Toolkit.
# 
# AUTHOR
#   Andy Wardley <abw@kfs.org>
#
# COPYRIGHT
#   Copyright (C) 1996-2001 Andy Wardley.  All Rights Reserved.
#   Copyright (C) 1998-2001 Canon Research Centre Europe Ltd.
#
#   This module is free software; you can redistribute it and/or
#   modify it under the same terms as Perl itself.
#
# TODO
#   * parse_interpolate()
#   * hash arrays with assignments
#   * names: filename + filename
#   * nameargs: names args | names ( args )
#
# REVISION
#   $Id: Parser.pm,v 1.2 2001/06/19 16:15:25 abw Exp $
#
#========================================================================

package Template::Parser;

require 5.005;

use strict;
use Template::Constants;
use vars qw( $VERSION $DEBUG $ERROR $ETYPE $TAG_STYLE $DEFAULT_STYLE );
use base qw( Template::Base );

$VERSION = sprintf("%d.%02d", q$Revision: 1.2 $ =~ /(\d+)\.(\d+)/);
$DEBUG   = 0 unless defined $DEBUG;
$ERROR   = '';
$ETYPE   = 'parser';			    # exception type

# temp hack
my $HANDLER = $Template::Factory::MODULES->{ parser_handler };

$TAG_STYLE   = {
    'default'   => [ '\[%',    '%\]'    ],
    'template'  => [ '\[%',    '%\]'    ],
    'template1' => [ '[\[%]%', '%[\]%]' ],
    'metatext'  => [ '%%',     '%%'     ],
    'html'      => [ '<!--',   '-->'    ],
    'mason'     => [ '<%',     '>'      ],
    'asp'       => [ '<%',     '%>'     ],
    'php'       => [ '<\?',    '\?>'    ],
    'star'      => [ '\[\*',   '\*\]'   ],
};

$DEFAULT_STYLE = {
    START_TAG   => $TAG_STYLE->{ default }->[0],
    END_TAG     => $TAG_STYLE->{ default }->[1],
    ANYCASE     => 0,
    INTERPOLATE => 0,
    PRE_CHOMP   => 0,
    POST_CHOMP  => 0,
    V1DOLLAR    => 0,
    EVAL_PERL   => 0,
};

my $REGEXEN = {
    number   => qr/ \G \s* ([+-]? \d+ (?:\.\d+)? ) /x,
    literal  => qr/ \G \s* ' ( (?:\\'|[^'])* ) ' /x,
    quoted   => qr/ \G \s* " ( (\\\\|\\"|.|\n)*? ) " /x,
    ident    => qr/ \G (\w+) /x,
    interp   => qr/ \G \$ (\w+) /x,
    embed    => qr/ \G \${ \s* ([^}]*?) \s* } /x,
    filename => qr/ \G ([\w\.\/:]+) /x,
    resource => qr/ \G (\w+):([\w\.\/:]+) /x,
    binop    => qr/ \G \s* ( [\-\+\*\/%] 
                           | [=!<>]= | [<>] | eq | ne | [lg][et] 
                           | &&? | \|\| | or | and  ) \s* /x,
    text     => qr/ \G ( (?: \\. | [^\$] )+ ) /x,
};

my $PARENS = {
    '(' => '\)',
    '[' => '\]',
    '<' => '\>',
    '|' => '\|',
};



#------------------------------------------------------------------------
# init(\%config)
#
# Initialiser method called by base class new() method.
#------------------------------------------------------------------------

sub init {
    my ($self, $config) = @_;

    # create parser compiler if not supplied
    $self->{ _COMPILER } = $config->{ COMPILER }
	|| $self->create( compiler => $config )
	|| return;

    # create default parser context
    $self->new_scope($config)
	 || return;

    return $self;
}


#========================================================================
# methods for parsing the basic Template Toolkit language
#========================================================================

#------------------------------------------------------------------------
# parse_template($text)
#------------------------------------------------------------------------

sub parse_template {
    my $self = shift;
    my $text = shift;
    my $options = ref $_[0] eq 'HASH' ? shift : { @_ };
    my $textref = (ref $text ? $text : \$text);

    $self->DEBUG('parse_template("', $self->_inspect_text($$textref), "\")\n")
	if $DEBUG;

    my $compiler = $self->{ _COMPILER } 
	|| return $self->throw('no compiler defined');

    $self->DEBUG("about to create handler\n");
    my $handler = $compiler->template_handler()
	|| return $self->throw($compiler->error());

    $self->push_handler($handler)
	|| return;

    $self->_tokenise($textref)
	|| return;

    return $handler->content();
}


#------------------------------------------------------------------------
# parse_block($text)
#------------------------------------------------------------------------

sub parse_block {
    my $self = shift;
    my $text = shift;
    my $options = ref $_[0] eq 'HASH' ? shift : { @_ };
    my $textref = ref $text ? $text : \$text;
}


#------------------------------------------------------------------------
# parse_statement($text, \%options)
#------------------------------------------------------------------------

sub parse_statement {
    my $self = shift;
    my $text = shift;
    my $options = ref $_[0] eq 'HASH' ? shift : { @_ };
    my $textref = ref $text ? $text : \$text;
    my ($stmt, $expr, @setlist);

    my $compiler = $self->{ _COMPILER } 
	|| return $self->throw('no compiler defined');

    # skip any leading whitespace
    $$textref =~ / \G \s* /gcx;

    if (defined ($stmt = $self->parse_assign($textref, pushback => 1, no_compile => 1))) {

	$self->DEBUG("stmt got an assign: [ $stmt->[0], $stmt->[1] ]\n")
	    if $DEBUG;

	while ($stmt) {
	    push(@setlist, $stmt);
	    $stmt = $self->parse_assign($textref, no_compile => 1);
	}
	return $compiler->setlist(\@setlist)
	    || $self->error($compiler->error());
    }
    elsif (defined ($stmt = $self->parse_expression($textref))) {
	$self->DEBUG("stmt got an expression: $stmt\n")
	    if $DEBUG;
	return $stmt;
    }
    else {
	return $self->parse_error("cannot parse statement");
    }

    return $stmt;
}


#------------------------------------------------------------------------
# parse_expression($text, \%options)
#------------------------------------------------------------------------

sub parse_expression {
    my $self = shift;
    my $text = shift;
    my $options = ref $_[0] eq 'HASH' ? shift : { @_ };
    my $textref = ref $text ? $text : \$text;
    my ($term, $expr, $var, $op, $value, $true, $false);

    my $compiler = $self->{ _COMPILER } 
	|| return $self->throw('no compiler defined');

    # skip any leading whitespace
    $$textref =~ / \G \s* /gcx;

    # ! expression
    if ($$textref =~ / \G ! \s* /gcx) {
	defined ($expr = $self->parse_expression($textref))
	    || return;
	$expr = "! $expr";
	return $expr;
    }
    # ( expression )
    elsif ($$textref =~ / \G \( \s* /gcx) {
	if (defined ($term = $self->parse_assign($textref, pushback => 1))) {
	    $self->DEBUG("expr got an assign: $term\n")
		if $DEBUG;
	    $self->throw('no terminating )')
		unless $$textref =~ / \G \s* \) /gcx;
	}
	elsif (defined ($expr = $self->parse_expression($textref))) {
	    $self->throw('no terminating )')
		unless $$textref =~ / \G \s* \) /gcx;
	    $term = "($expr)";
	}
    }
    elsif (defined ($term = $self->parse_term($textref))) {
	$self->DEBUG("expr got a term: $term\n")
	    if $DEBUG;
    }
    else {
	return undef;
    }

    # expr BINOP expr
    if ($$textref =~ /$REGEXEN->{ binop }/gc) {
	$op = $1;
	$self->DEBUG("expr got a binop: $op\n")
	    if $DEBUG;

	defined ($expr = $self->parse_expression($textref))
	    || $self->throw("missing expression after '$op'");
	
	$expr = "$term $op $expr";
    }
    # expr ? expr : expr
    elsif ($$textref =~ / \G \s* \? \s* /gcx) {
	$self->DEBUG("expr got ? :\n")
	    if $DEBUG;
	
	defined ($true = $self->parse_expression($textref))
	    || $self->throw("missing expression after '?'");
	
	$self->throw("missing ':' after '?' expression")
	    unless $$textref =~ / \G \s* : \s* /gcx;
	
	defined ($false = $self->parse_expression($textref))
	    || $self->throw("missing expression after ':'");
	
	$expr = "$term ? $true : $false";
    }
    else {
	$expr = $term;
    }

    return $expr;
}	


#------------------------------------------------------------------------
# parse_assign($self, $text)
#
# Parse a TT variable assignment.  e.g. foo.bar(3).baz = x.y.z
#------------------------------------------------------------------------

sub parse_assign {
    my $self = shift;
    my $text = shift;
    my $options = ref $_[0] eq 'HASH' ? shift : { @_ };
    my $textref = ref $text ? $text : \$text;
    my ($var, $op, $value, $assign);

    my $compiler = $self->{ _COMPILER } 
	|| return $self->throw('no compiler defined');

    # save string position in case we need to backtrack
    my $pos = pos $$textref;

    if (defined ($var = $self->parse_variable($textref))) {
	if ($$textref =~ / \G \s* (=>?) \s* /gcx) {
	    $op = $1;

	    $self->DEBUG("expr got assignment\n")
		if $DEBUG;

	    defined ($value = $self->parse_expression($textref))
		|| $self->throw("missing expression after $op");

	    if ($options->{ no_compile }) {
		# we want to compile SET directives with multiple
		# assignments into one call to $stash->set() so 
		# this option allows the caller (parse_statement())
		# to collect all assignments uncompiled and then
		# compile them in one go
		$self->DEBUG("parse_assign() returning [ $var, $value ]\n")
		    if $DEBUG;
		$assign = [ $var, $value ];
	    }
	    else {
		# otherwise go ahead and compile it
		defined ($assign = $compiler->assign($var, $value))
		    || $self->throw($compiler->error());
		$self->DEBUG("parse_assign() returning compiled '$assign'\n")
		    if $DEBUG;
	    }
	    return $assign;
	}
	else {
	    # we check for following whitespace to ensure this really can
	    # be passed off as a variable and not something like qw[ 
	    # where 'qw' has been matched but shouldn't have
	    if ($options->{ pushback } && $$textref =~ / \G \s+ /gcx) {	
		# cache variable
#		$self->DEBUG("not an assign, CACHING\n")
#		    if $DEBUG;
		$self->{ _PUSHBACK_VAR } = $var;
	    }
	    else {
		# rewind string position (backtrack)
#		$self->DEBUG("not an assign, REWINDING\n")
#		    if $DEBUG;
		pos $$textref = $pos;
	    }
	}
    }
    return undef;
}


#------------------------------------------------------------------------
# parse_embedded($self, $text)
#
# Parse a directive embedded within ${ ... }.
#------------------------------------------------------------------------

sub parse_embedded {
    my $self = shift;
    my $text = shift;
    my $options = ref $_[0] eq 'HASH' ? shift : { @_ };

    my $compiler = $self->{ _COMPILER } 
	|| return $self->throw('no compiler defined');

    my $handler = $self->{ _HANDLER_STACK }->[-1]
	|| return $self->error('no current parser handler');

    # clone the top handler so that we have a clean CONTENT
    # buffer to work with
    my $clone = $handler->clone();
    $clone->directive($self, $text)
	|| return $self->parse_error($clone->error(), $text);
    
    return $compiler->block($clone->content())
	|| $self->error($compiler->error());
}



#------------------------------------------------------------------------
# parse_term($text, \%options)
#
# Parse a term which can be a number, C<5>, a literal string, C<'foo'>, a 
# quoted string, C<"foo$bar">, a variable, C<foo.bar(x)>, a list,
# C<[ foo, bar.baz, "hello" ]> or a hash C<{ name => 'Fred', ... }>
#------------------------------------------------------------------------

sub parse_term {
    my $self = shift;
    my $text = shift;
    my $options = ref $_[0] eq 'HASH' ? shift : { @_ };
    my $textref = ref $text ? $text : \$text;
    my $compiler = $self->{ _COMPILER } 
	|| return $self->throw('no compiler defined');
    my ($term, $args);

    # check variable cache for any variable read-ahead (e.g. 
    # by looking ahead for an '=' assignment)
    if ($self->{ _PUSHBACK_VAR } 
	&& defined ($term = $self->parse_variable($textref))) {
	defined ($term = $compiler->variable($term))
	    || $self->throw($compiler->error());
	$self->DEBUG("got pushback variable: $term\n") if $DEBUG;
	return $term;
    }

    # term can be an number such as: 5...
    if ($$textref =~ /$REGEXEN->{ number }/gc) {
	$self->DEBUG("got number: $1\n") if $DEBUG;
	defined($term = $compiler->number($1))
	    || return $self->error($compiler->error());
    }
    # ...or a literal (single quoted) string like: 'foo'...
    elsif ($$textref =~ /$REGEXEN->{ literal }/gc) {
	$self->DEBUG("got literal: $1\n") if $DEBUG;
	defined($term = $compiler->literal($1))
	    || return $self->error($compiler->error());
    }
    # ...or a double quoted string like: "foo$bar"...
    elsif ($$textref =~ /$REGEXEN->{ quoted }/gc) {
	$self->DEBUG("got quoted: $1\n") if $DEBUG;
	defined($term = $self->parse_interpolate($1))
	    || return;
    }
    # ...or a list: [ 1, two, '3' ]
    elsif ($$textref =~ / \G \[ \s* /gcx) {
	$term = $self->parse_list($textref);
	$self->DEBUG("got list: @$term\n") if $DEBUG;
	$self->throw("terminating ']' missing after list definition")
	    unless $$textref =~ / \G \s* \] /gcx;
	defined($term = $compiler->list($term))
	    || return $self->error($compiler->error());
    }
    # ...or a quote list: qw[ foo bar baz ]
    elsif ($$textref =~ / \G qw([\[\(\{\<\|]) \s* /gcx) {
	my $paren = $PARENS->{ $1 };
	if ($$textref =~ / \G ([^$paren]*) $paren /gcx) {
	    $term = $1;
	    $term =~ s/\s+$//;
	    $self->DEBUG("got qwlist: $term\n") if $DEBUG;
	    defined($term = $compiler->qwlist($term))
		|| return $self->error($compiler->error());
	}
	else {
	    $self->throw("cannot parse qw[] list");
	}
#	    $self->throw("terminating ']' missing after qw[ ] list definition")
#		unless $$textref =~ / \G \s* \] /gcx;

    }
    # ...or a hash: { name => 'Fred', email => myorg.email('fred') }
    elsif ($$textref =~ / \G \{ \s* /gcx) {
	$term = $self->parse_hash($textref);
	$self->DEBUG("got hash: @$term\n") if $DEBUG;
	$self->throw("terminating '}' missing after hash definition")
	    unless $$textref =~ / \G \s* \} /gcx;
	defined($term = $compiler->hash($term))
	    || return $self->error($compiler->error());
    }
    # ...or a resource: http://www.example.com/blah.xml
    elsif ($$textref =~ /$REGEXEN->{ resource }/gc) {
	$self->DEBUG("got resource: ($1) ($2)\n") if $DEBUG;
	return "$1_resource($2)";
    }
    # ...or a variable like: foo.bar(baz)
    elsif (defined ($term = $self->parse_variable($textref))) {
	defined ($term = $compiler->variable($term))
	    || $self->throw($compiler->error());
	$self->DEBUG("got variable: $term\n") if $DEBUG;
    }
    # ...or it's not a term
    else {
	return undef;
    }

    return $term;
}


#------------------------------------------------------------------------
# parse_variable($self, $text)
#
# Parse a TT variable which can be anything from a simple, single node
# variable like 'foo' to a complex, multi-node variable with deeply
# nested arguments lists such as 'foo(x.y(z(a)), b).bar("hello").baz(x)'.
#------------------------------------------------------------------------

sub parse_variable {
    my $self = shift;
    my $text = shift;
    my $options = ref $_[0] eq 'HASH' ? shift : { @_ };
    my $textref = ref $text ? $text : \$text;

    if (my $pushback = $self->{ _PUSHBACK_VAR }) {
	$self->{ _PUSHBACK_VAR } = undef;
#	$self->DEBUG("parse_variable() returning pushed back variable: ",
#		     $self->_inspect_ref($pushback), "\n")
#	    if $DEBUG;
	return $pushback;
    }
    my $compiler = $self->{ _COMPILER } 
	|| return $self->throw('no compiler defined');
    my ($node, @nodes);

    # skip any leading whitespace
    $$textref =~ / \G \s* /gcx;

    # parse the first node
    $node = $self->parse_node($textref);
    return undef unless defined $node;
    push(@nodes, @$node);

    # look for a following '.' then expect another node
    while ($$textref =~ / \G \.(?!\.) /gcx) {
	$node = $self->parse_node($textref);
	$self->throw("expected node after '.'")
	    unless defined $node;
	push(@nodes, @$node);
    }

    return \@nodes;
}


#------------------------------------------------------------------------
# parse_node($self, \$text)
#
# Parse a TT variable node containing an identifier with an optional 
# argument list in parentheseis, e.g. 'foo' or 'foo(x, y, z(a))'.
#------------------------------------------------------------------------
 
sub parse_node {
    my $self = shift;
    my $text = shift;
    my $options = ref $_[0] eq 'HASH' ? shift : { @_ };
    my $textref = ref $text ? $text : \$text;
    my $compiler = $self->{ _COMPILER } 
	|| return $self->throw('no compiler defined');
    my ($ident, $args);

    # node can be a number like '4'...
    if ($$textref =~ /$REGEXEN->{ number }/gc) {
	$self->DEBUG("got number: $1\n") if $DEBUG;
	defined($ident = $compiler->number($1))
	    || return $self->throw($compiler->error());
    }
    #  ...or an identifier like 'foo'...
    elsif ($$textref =~ /$REGEXEN->{ ident }/gc) {
	$self->DEBUG("got ident: $1\n") if $DEBUG;
	defined($ident = $compiler->literal($1))
	    || return $self->throw($compiler->error());
    }
    # ...or an interpolated value like '$foo'...
    elsif ($$textref =~ /$REGEXEN->{ interp }/gc) {
	$self->DEBUG("got interp: $1\n") if $DEBUG;
	defined($ident = $self->parse_variable($1)) 
	    || $self->throw('invalid variable reference');
	defined($ident = $compiler->variable($ident))
	    || $self->throw($compiler->error());
    }
    # ...or an embedded directive like '${ foo.bar(x) }'...
    elsif ($$textref =~ /$REGEXEN->{ embed }/gc) {
	$self->DEBUG("got embed: $1\n") if $DEBUG;
	defined($ident = $self->parse_embedded($1)) 
	    || $self->throw('invalid embedded term');
    }
    # ...or it's not a node
    else {
	return undef;
    }

    # does it have an argument list?
    if ($$textref =~ / \G \( /gcx) {
	$args = $self->parse_args($textref);

	$self->throw("expected ')' to terminate argument list but got '"
		     . $self->_inspect_text($self->next_text($textref)) . "'")
	    unless $$textref =~ / \G \s* \) /gcx;

	defined ($args = $compiler->args($args))
	    || $self->throw($compiler->error());
    }
    $args = 0 unless $args;
    return [ $ident, $args ];
}


#------------------------------------------------------------------------
# parse_list($self, \$text)
#
# Parse a list of arguments, each of which can be a TT term separated
# by comma and/or whitespace, e.g. 'foo, "bar", baz' or 'x y(z)'
#------------------------------------------------------------------------

sub parse_list {
    my ($self, $text) = @_;
    my $textref = ref $text ? $text : \$text;
    my ($arg, @args);

    # skip any leading whitespace
    $$textref =~ / \G \s* /gcx;

    $arg = $self->parse_expression($textref);
    return undef unless defined $arg;
    push(@args, $arg);

    while ($$textref =~ / \G (?: ( \s* \.\. \s* )
			       | ( \s* , \s* )
			       | ( \s+ )
			      )
			      /gcx) {
	last unless defined ($arg = $self->parse_expression($textref));
	if ($1) {
	    $self->DEBUG("list got range: $args[-1] .. $arg\n")
		if $DEBUG;
	    push(@args, [pop @args, $arg]);
	}
	else {
	    $self->DEBUG("list got item: $arg\n")
		if $DEBUG;
	    push(@args, $arg);
	}
    }
    return \@args;
}


#------------------------------------------------------------------------
# parse_hash($self, \$text)
#
# Parse the content of a hash array definition.
#------------------------------------------------------------------------

sub parse_hash {
    my ($self, $text) = @_;
    my $textref = ref $text ? $text : \$text;
    my ($key, $value, $op, @hash);
    my $compiler = $self->{ _COMPILER } 
	|| return $self->throw('no compiler defined');

    while ( ($$textref =~ /$REGEXEN->{ ident   }/gcx)
	 || ($$textref =~ /$REGEXEN->{ literal }/gcx) ) {
	$key = $1;
	if ($$textref =~ / \G \s* (=>?) \s* /gcx) {
	    $op = $1;
	    defined ($value = $self->parse_expression($textref))
		|| $self->throw("missing expression after '$op'");
	    defined ($key = $compiler->literal($key))
		|| $self->throw($compiler->error());
	    push(@hash, $key, $value);
	    
	    # skip comma and/or whitespace
	    $$textref =~ / \G (?:\s*,)? \s+ /gcx;
	}
	else {
	    $self->throw("missing '=>' after '$key'");
	}
    }

    return \@hash;
}


#------------------------------------------------------------------------
# parse_args($text)
#
# Parses argument list which can contain regular terms (expr?) or named
# parameters.  Simple items are pushed to the @$args list, named params 
# are pushed into the sub-list in element 0 of $args.
#------------------------------------------------------------------------

sub parse_args {
    my $self = shift;
    my $text = shift;
    my $options = ref $_[0] eq 'HASH' ? shift : { @_ };
    my $textref = ref $text ? $text : \$text;
    my $compiler = $self->{ _COMPILER } 
	|| return $self->throw('no compiler defined');
    my ($pos, $op, $key, $value);
    my $args = [ [ ] ];

    # skip any leading whitespace
    $$textref =~ / \G \s* /gcx;

    $self->DEBUG("parse_args [$$textref]\n")
	if $DEBUG;

    while (1) {
	# save current string position
	$pos = pos $$textref;

	# look for something than be can be the LHS of an assignment
	if ( ($$textref =~ /$REGEXEN->{ ident   }/gcx) ||
	     ($$textref =~ /$REGEXEN->{ literal }/gcx) ) {
	    $key = $1;

	    $self->DEBUG("got ident or literal: $key\n")
		if $DEBUG;
	    
	    if ($$textref =~ / \G \s* (=>?) \s* /gcx) {
		# if it's followed by '=' or '=>' then it's a named param
		$op = $1;
		defined ($value = $self->parse_expression($textref))
		    || $self->throw("missing expression after $op");
		defined ($key = $compiler->literal($key))
		    || $self->throw($compiler->error());
		push(@{$args->[0]}, $key, $value);
		$self->DEBUG("pushed [$key] => [$value]\n") if $DEBUG;
		next;				## MATCHED NAMED PARAM ##
	    }
	    else {
		# otherwise it must have been a simple term so reset
		# string position and drop-through
		pos $$textref = $pos;
	    }
	}

	if (defined ($value = $self->parse_expression($textref))) {
	    $self->DEBUG("got value: [$value]\n") if $DEBUG;
	    push(@$args, $value);		## MATCHED SIMPLE VALUE ##
	}
	else {
	    $self->DEBUG("no term, returning: [@$args]\n") if $DEBUG;
	    last;				## ALL DONE ##
	}
    }
    continue {
	# skip comma and/or whitespace
	$$textref =~ / \G (?:\s*,)? \s+ /gcx;
    }

    # trailing comma/whitespace
    $$textref =~ / \G (?:\s*,)? \s* /gcx;

    return $args;
}
    

#------------------------------------------------------------------------
# parse_filename($self, \$text)
#
# Parse a template filename.
#------------------------------------------------------------------------

sub parse_filename {
    my $self = shift;
    my $text = shift;
    my $options = ref $_[0] eq 'HASH' ? shift : { @_ };
    my $textref = ref $text ? $text : \$text;
    my $compiler = $self->{ _COMPILER } 
	|| return $self->throw('no compiler defined');
    my $filename;

    # naked filename
    if ($$textref =~ /$REGEXEN->{ filename }/gc) {
	$self->DEBUG("got naked filename: $1\n") if $DEBUG;
	defined ($filename = $compiler->literal($1))
	    || $self->throw($compiler->error());
    }
    # 
    elsif ($$textref =~ /$REGEXEN->{ quoted }/gc) {
	$self->DEBUG("got quoted filename: \"$1\"\n") if $DEBUG;
	defined($filename = $self->parse_interpolate($1))
	    || return;
    }
    #
    elsif ($$textref =~ /$REGEXEN->{ literal }/gc) {
	$self->DEBUG("got literal filename: '$1'\n") if $DEBUG;
	defined ($filename = $compiler->literal($1))
	    || $self->throw($compiler->error());
    }
    elsif ($$textref =~ / \G \$ /gcx) {
	defined ($filename = $self->parse_variable($1))
	    || return;
	defined ($filename = $compiler->variable($filename))
	    || $self->throw($compiler->error());
    }

    return $filename;
}


#------------------------------------------------------------------------
# parse_interpolate($self, $text)
#
# Parses the contents (i.e. not the enclosing '"' characters) of a double
# quoted string which may contain embedded variables and/or directives.
#------------------------------------------------------------------------

sub parse_interpolate {
    my $self = shift;
    my $text = shift;
    my $options = ref $_[0] eq 'HASH' ? shift : { @_ };
    my $textref = ref $text ? $text : \$text;

    my $compiler = $self->{ _COMPILER } 
	|| return $self->throw('no compiler defined');

    # TODO
    my @tokens  = ();
    my $var;

    while ( $$textref =~ / 
	    ( (?: \\. | [^\$] )+ )     # escaped or non-'$' character
	    | \${ \s* ([^}]*?) \s* }   # ${ DIRECTIVE }
	    | \$ (\w+)
    /gx ) {

	# preceding text
	if ($1) {
	    push(@tokens, "\"$1\"");
	}
	# $variable reference
        elsif ($2) {
	    push(@tokens, $self->parse_embedded($2) || return);
	}
	else {
	    push(@tokens, $self->parse_term($3) || return);
	}
    }

    return $compiler->quoted(\@tokens)
	|| $self->error($compiler->error());
}



#========================================================================
# Methods to update parser options in current scope
#========================================================================

#------------------------------------------------------------------------
# tags($newtags)
#
# Set tag style, e.g. 'star', or start/end tags separated by whitespace,
# e.g. '[* *]'.
#------------------------------------------------------------------------

sub tags {
    my ($self, $text) = @_;
    my @tags = split(/\s+/, $text);
    my ($tags, $stag, $etag);
    
    $self->DEBUG($self->ID, "->tags($text)\n")
	if $DEBUG;

    if (scalar @tags > 1) {
	($stag, $etag) = map { quotemeta($_) } @tags;
    }
    elsif ($tags = $TAG_STYLE->{ $tags[0] }) {
	($stag, $etag) = @$tags;
    }
    else {
	warn "invalid TAGS style: $tags[0]\n";
    }

    my $scope = $self->{ _SCOPE_STACK }->[-1]
	|| return $self->error('no current parser scope (tags)');

    @$scope{ qw( START_TAG END_TAG ) } = ($stag, $etag);
    $scope->{ REGEX } = $self->_tokenise_regex();
}


#------------------------------------------------------------------------
# interpolate($flag)
#
# Set interpolate mode on ($flag == true) or off ($flag == false).
#------------------------------------------------------------------------

sub interpolate {
    my ($self, $flag) = @_;
    my $inc = $flag ? +1 : -1;

    my $scope = $self->{ _SCOPE_STACK }->[-1]
	|| return $self->error('no current parser scope (interpolate)');

    $scope->{ INTERPOLATE } += $inc;
    $self->DEBUG($self->ID, "->interpolate($flag) => $scope->{ INTERPOLATE }\n")
	if $DEBUG;
    $scope->{ REGEX } = $self->_tokenise_regex();
}



#========================================================================
# Methods to manipulate handlers 
#========================================================================

#------------------------------------------------------------------------
# push_handler($handler)
#
# Push a new handler onto the top of the handler stack.
#------------------------------------------------------------------------

sub push_handler {
    my ($self, $handler) = @_;
    my $handlers = $self->{ _HANDLER_STACK } ||= [ ];
    push(@$handlers, $handler);
}


#------------------------------------------------------------------------
# pop_handler()
#
# Pop top handler from the handler stack.
#------------------------------------------------------------------------

sub pop_handler {
    my $self = shift;
    my $handlers = $self->{ _HANDLER_STACK };

    return $self->error('no current parser handler')
	unless @$handlers;

    return pop(@$handlers);
}


#------------------------------------------------------------------------
# new_handler(@args)
#
# Create new handler by calling the new_handler() method on the top
# handler.
#------------------------------------------------------------------------

sub new_handler {
    my $self = shift;
    my $handlers = $self->{ _HANDLER_STACK };

    return $self->error('no current parser handler')
	unless @$handlers;

    my $top = $handlers->[-1];
    my $new = $top->new_handler(@_)
	|| return $self->error($top->error());

    push(@$handlers, $new);
    return $new;
}

#------------------------------------------------------------------------
# current_handler()
#
# Return top handler from stack without removing it
#------------------------------------------------------------------------

sub current_handler {
    my $self = shift;
    return $self->{ _HANDLER_STACK }->[-1]
	|| $self->error('no current parser handler');
}


#------------------------------------------------------------------------
# old_handler()
#
# Pop top handler from the handler stack then call the old_handler()
# method on the item directly below it, passing itself as a reference.
#------------------------------------------------------------------------

sub old_handler {
    my $self = shift;
    my $handlers = $self->{ _HANDLER_STACK };

    return $self->error('no current parser handler')
	unless @$handlers > 1;

    my $old = pop @$handlers;
    my $top = $handlers->[-1];

    $top->old_handler($old)
	|| return $self->error($old->error());

    return $top;
}


#========================================================================
# Methods to manipulate scopes (current parser configuration)
#========================================================================

#------------------------------------------------------------------------
# push_scope($scope)
#
# Push a new scope onto the top of the scope stack.
#------------------------------------------------------------------------

sub push_scope {
    my ($self, $scope) = @_;
    my $stack = $self->{ _SCOPE_STACK } ||= [ ];
    push(@$stack, $scope);
}


#------------------------------------------------------------------------
# pop_scope()
#
# Pop top scope from the scope stack.
#------------------------------------------------------------------------

sub pop_scope {
    my $self = shift;
    my $stack = $self->{ _SCOPE_STACK };

    return $self->error('no current parser scope (pop)')
	unless @$stack;

    return pop(@$stack);
}

#------------------------------------------------------------------------
# new_scope(\%args)
#
# Create a new scope by merging the existing scope configuration with 
# new configuration arguments passed by hash reference or as a list.
#------------------------------------------------------------------------

sub new_scope {
    my $self  = shift;
    my $args  = ref($_[0]) eq 'HASH' ? shift : { @_ };
    my $stack = $self->{ _SCOPE_STACK } ||= [ ];
    my $scope = { %{ $stack->[-1] || $DEFAULT_STYLE } };
    my ($tagstyle, $tags, $start, $end, $key, $value);

    # expand START_TAG and END_TAG from specified TAG_STYLE
    if ($tagstyle = $args->{ TAG_STYLE }) {
	return $self->error("Invalid tag style: $tagstyle")
	    unless defined ($tags = $TAG_STYLE->{ $tagstyle });
	($start, $end) = @$tags;
	$args->{ START_TAG } ||= $start;
	$args->{   END_TAG } ||= $end;
    }

    $self->DEBUG($self->ID, " new_scope(", 
		 $self->_inspect_hash($args), ")\n")
	if $DEBUG;

    foreach $key (keys %$DEFAULT_STYLE) {
	$scope->{ $key } = $value,
	    if defined ($value = $args->{ $key });
    }
    push(@$stack, $scope);
    return $scope;
}


#------------------------------------------------------------------------
# current_scope()
#
# Return the current scope, that is a reference to a hash containing
# current parser configuration.
#------------------------------------------------------------------------

sub current_scope {
    my $self = shift;
    return $self->{ _SCOPE_STACK }->[-1]
	|| $self->error('no current parser context');
}


#------------------------------------------------------------------------
# old_scope()
#
# Pop the top (current) scope off the scope stack and return the item
# below it.
#------------------------------------------------------------------------

sub old_scope {
    my $self  = shift;
    my $stack = $self->{ _SCOPE_STACK };
    return $self->error('parser context stack corrupt')
	unless @$stack > 1;
    pop @$stack;
    return $stack->[-1];
}



#========================================================================

#------------------------------------------------------------------------
# _tokenise($self, $textref)
#------------------------------------------------------------------------

sub _tokenise {
    my ($self, $textref) = @_;
    my ($pre, $prelines, $dir, $dirlines, $linespan, $postlines);
    my ($var, $bracevar, $chomp);
    my ($scope, $handler);
    my $line = 1;

    return $self->throw('no input text')
	unless defined $textref;

    $self->DEBUG($self->ID, "->_tokenise('",
		 $self->_inspect_text($$textref), "')\n")
	if $DEBUG;

    $scope = $self->{ _SCOPE_STACK }->[-1]
	|| return $self->error('no current parser context');

    $handler = $self->{ _HANDLER_STACK }->[-1]
	|| return $self->error('no current parser handler');

    $self->{ _LINE }     = \$line;
    $self->{ _LINESPAN } = \$linespan;
    my ($stag, $etag) = @$scope{ qw( START_TAG END_TAG ) };

    $scope->{ REGEX } = $self->_tokenise_regex();

    while ($$textref =~ s/$scope->{ REGEX }//s) {
	($pre, $dir, $var, $bracevar) = ($1, $2, $3, $4);
	$pre = '' unless defined $pre;
	$dir = '' unless defined $dir;
    
	$postlines = $dirlines = 0;
	$prelines  = ($pre =~ tr/\n//);      # NULL - count only
	$dirlines  = ($dir =~ tr/\n//);      # ditto

	$scope = $self->{ _SCOPE_STACK }->[-1]
	    || return $self->error('no current parser scope');

	$handler = $self->{ _HANDLER_STACK }->[-1]
	    || return $self->error('no current parser handler');

	# the directive CHOMP options may modify the preceding text
	for ($dir) {
	    # remove leading whitespace and check for a '-' chomp flag
	    s/^([-+\#])?\s*//s;
	    if ($1 && $1 eq '#') {
		# comment out entire directive
		$dir = '';
	    }
	    else {
		my $prechomp = $scope->{ PRE_CHOMP };
		$chomp = ($1 && $1 eq '+') ? 0 : ($1 || $prechomp);
		my $space = $prechomp == &Template::Constants::CHOMP_COLLAPSE ? ' ' : '';

		# chomp off whitespace and newline preceding directive
		$chomp and $pre =~ s/(\n|^)([ \t]*)\Z/($1||$2) ? $space : ''/me;
	    }

	    # remove trailing whitespace and check for a '-' chomp flag
	    s/\s*([-+])?\s*$//s;
	    my $postchomp = $scope->{ POST_CHOMP };
	    $chomp = ($1 && $1 eq '+') ? 0 : ($1 || $postchomp);
	    my $space = $postchomp== &Template::Constants::CHOMP_COLLAPSE ? ' ' : '';

	    # only chomp newline if it's not the last character
	    $chomp and $$textref =~ s/^([ \t]*)\n(.|\n)/(($1||$2) ? $space : '') . $2/e
		and $postlines++;
	}

	# notify handler of text preceding directive
	$handler->text($self, $pre)
	    || return $self->parse_error($handler->error(), $pre);

	# increment line number to start of directive
        $line += $prelines;
	$linespan = $dirlines;

	if (length $dir) {
	    $handler->directive($self, $dir)
		|| return $self->parse_error($handler->error(), $dir);
	}
	elsif (defined $var) {
	    $handler->directive($self, $var)
		|| return $self->parse_error($handler->error(), $dir);
	}
	elsif (defined $bracevar) {
	    $handler->directive($self, $bracevar)
		|| return $self->parse_error($handler->error(), $bracevar);
	}	    

	# increment line counter to include directive lines and any extra
	# newline chomped off the start of the following text
	$line += $dirlines + $postlines;
	$linespan = 0;
    }

    # notify handler of text at end of document
    $handler->text($self, $$textref)
	if length $$textref;

    return 1;
}


#------------------------------------------------------------------------
# _tokenise_regex()
#
# Generate regex for tokenising input text according to current parser
# tags style, interpolate flag, etc.
#------------------------------------------------------------------------

sub _tokenise_regex {
    my $self = shift;

    my $scope = $self->{ _SCOPE_STACK }->[-1]
	|| return $self->throw('no current parser scope');

    my ($stag, $etag, $interp) = 
	@$scope{ qw( START_TAG END_TAG INTERPOLATE ) };

    if ($interp) {
	return qr/
	    ^( (?: \\. | [^\$] )+? )	    # $1 - preceding text
	     (?: $stag			    # start of tag
		 (.*?)			    # $2 - tag contents
		 $etag			    # end of tag
	     |   \$([\w\.]+)		    # $3 - $foo.bar
	     |   \$\{ \s* ([^\}]*?) \s* \}  # $4 - ${ foo.bar }
	     )
	/sx;
    }
    else {
	return qr/
	    ^(.*?)			    # $1 - preceding text
	     (?: $stag			    # start of tag
		 (.*?)			    # $2 - tag contents
		 $etag			    # end of tag
	     )
	/sx;
    }	    
}


#------------------------------------------------------------------------
# location()
#
# Return current parse location in terms of a file number or 'n-m' range.
#------------------------------------------------------------------------

sub location {
    my $self = shift;
    my ($line, $linespan) = @$self{ qw( _LINE _LINESPAN ) };
    $line = ref($line) ? $$line : $line;
    $linespan = ref($linespan) ? $$linespan : $linespan;
    unless (defined $line) {
	$line = 'unknown';
    }
    elsif ($linespan) {
	$line .= "-" . ($line + $linespan);
    }
    return $line;
}


#------------------------------------------------------------------------
# parse_error($msg, $dirtext)
#
# Method used to handle errors encountered during the parse process
# in the parse() method.  
#------------------------------------------------------------------------

sub parse_error {
    my ($self, $msg, $dirtext) = @_;
    if (defined $dirtext) {
	my $scope = $self->{ _SCOPE_STACK }->[-1];
	my ($stag, $etag) = @$scope{ qw( START_TAG END_TAG ) };
	$msg .= "\n  $stag $dirtext $etag";
    }

    $self->DEBUG("parse_error([$msg], [$dirtext]\n")
	if $DEBUG;

    return $self->error("line ", $self->location(), ": $msg");
}

sub parse_warning {
    my ($self, $msg, $dirtext) = @_;
    if (defined $dirtext) {
	my $scope = $self->{ _SCOPE_STACK }->[-1];
	my ($stag, $etag) = @$scope{ qw( START_TAG END_TAG ) };
	$msg .= "\n  $stag $dirtext $etag";
    }

    $self->DEBUG("parse_warning([$msg])\n")
	if $DEBUG;

    print STDERR "WARNING at line ", $self->location(), ": $msg\n";
}



sub next_text {
    my ($self, $textref) = @_;
    my $pos = pos $$textref;
    $$textref =~ / \G \s* (.*) /gcx;
    pos $$textref = $pos;
    return $1;
}


#========================================================================
# debug/housekeeping methods
#========================================================================

#------------------------------------------------------------------------
# ID
#
# Return a useful string describing $self for debugging purposes.
#------------------------------------------------------------------------

sub ID {
    my $self = shift;
    return "parser";
}


1;

__END__

