#============================================================= -*-Perl-*-
#
# Template::Parser::Handler
#
# DESCRIPTION
#   Module for handling the parsing of a chunk of template text.
#   Defines a set of directives which are valid within its scope.
#   These index to handlers in the Template::Compiler object which
#   created it.  The Handler accepts text() messages from the parser
#   which are pushed onto an internal CONTENT list, and directive()
#   messages which are routed to the appropriate DIRECTIVES handler
#   for compilation, the result of which is also pushed onto the 
#   CONTENT.
# 
# AUTHOR
#   Andy Wardley <abw@kfs.org>
#
# COPYRIGHT
#   Copyright (C) 1996-2001 Andy Wardley.  All Rights Reserved.
#   Copyright (C) 1998-2001 Canon Research Centre Europe Ltd.
#
#   This module is free software; you can redistribute it and/or
#   modify it under the same terms as Perl itself.
#
# REVISION
#   $Id: Handler.pm,v 1.2 2001/06/19 16:15:25 abw Exp $
#
#========================================================================

package Template::Parser::Handler;

require 5.005;

use strict;
use vars qw( $VERSION $DEBUG $ERROR $ETYPE );
use base qw( Template::Base );

$VERSION = sprintf("%d.%02d", q$Revision: 1.2 $ =~ /(\d+)\.(\d+)/);
$DEBUG   = 0 unless defined $DEBUG;
$ETYPE   = 'parser_handler';	# exception type
$ERROR   = '';


#------------------------------------------------------------------------
# init(\%config)
#
# Initialiser method called by base class new() method.
#------------------------------------------------------------------------

sub init {
    my ($self, $config) = @_;
    my $dirs = $config->{ DIRECTIVES } || { };
    my $class = ref $self;
    my $val;

    {	no strict 'refs';
	my $pkgdirs = \%{"$class\::DIRECTIVES"};
	$dirs = { %$pkgdirs, %$dirs } if %$pkgdirs;
    }

    # map any scalar directives (method names) to [ $self, $method ]
    $self->{ DIRECTIVES } = { 
	map { $val = $dirs->{ $_ }; ($_, ref $val ? $val : [ $self, $val ]) }
	keys %$dirs,
    };

    $self->{ CONTENT } = [ ];

    return $self;
}


#------------------------------------------------------------------------
# clone( DIRECTIVES => { DIRNAME => [ $obj, $method, \@args ], ... } )
#
# Create a clone of the current handler with the existing set of defined
# DIRECTIVES supplemented (and possibly masked by) any new directives
# passed as arguments.
#------------------------------------------------------------------------

sub clone {
    my $self = shift;
    my $args = ref $_[0] eq 'HASH' ? shift : { @_ };
    my $newdirs = $args->{ DIRECTIVES } || $args;	# NOTE
    my $olddirs = $self->{ DIRECTIVES } || { };
    $args->{ DIRECTIVES } = { %$olddirs, %$newdirs };
    my $class = ref $self;
    return $class->new($args);
}


#------------------------------------------------------------------------
# directives()
#
# Return internal hash defining directives with known compiler handlers.
#------------------------------------------------------------------------

sub directives {
    my $self = shift;
    return $self->{ DIRECTIVES };
}


#------------------------------------------------------------------------
# content()
#
# Return internal list containing compiled template content collected
# by this handler.
#------------------------------------------------------------------------

sub content {
    my $self = shift;
    return $self->{ CONTENT };
}


#------------------------------------------------------------------------
# text($parser, $text)
#
# Callback raised by the Template::Parser object when a chunk of text
# is identified in the source template.  Simply shifts the text onto 
# the internal CONTENT list.
#------------------------------------------------------------------------

sub text {
    my ($self, $parser, $text) = @_;
    my $line = $parser->location();
    $self->DEBUG($self->ID, "->text('", $self->_inspect_text($text), "') at line $line\n")
	if $DEBUG;
    push(@{ $self->{ CONTENT } }, $text);
    return 1;
}


#------------------------------------------------------------------------
# directive($parser, $directive)
#
# Callback raised by the Template::Parser object when a directive
# is identified in the source template.  Looks for a leading KEYWORD
# which matches a handler defined in the DIRECTIVES hash and calls
# the handler if found.  Otherwise, the directive is parsed as a
# statement and the result pushed onto the internal CONTENT list.
#------------------------------------------------------------------------

sub directive {
    my ($self, $parser, $dir) = @_;
    my ($keyword, $handler, $object, $method, $args, $result, $gap);

    my $line = $parser->location();
    $self->DEBUG($self->ID, "->directive('", $self->_inspect_text($dir), "') at line $line\n")
	if $DEBUG;

    while ($dir) {
	if ($dir =~ /^\s*(\w+)(\s*)/) {
	    $keyword = $1;
	    $gap = $2;
	    $self->DEBUG($self->ID, "matched directive [$keyword]\n")
		if $DEBUG;
	    
	    # need to think about UPPER/lower case conversion???
    
	    # look for handler
	    if ($handler = $self->{ DIRECTIVES }->{ $keyword }) {

		# handler may be \&coderef or [ $obj, $method, \@args ]
		if (ref $handler eq 'ARRAY') {
		    ($object, $method, $args) = @$handler;
		    $args ||= [ ];
		    
		    $self->DEBUG($self->ID, "about to call object $object method $method\n")
			if $DEBUG;
		    
		    eval {
			$result = $object->$method($parser, \$dir, @$args);
			return $self->error($object->error())
			    unless defined $result;
		    };
		    return $self->error($@) if $@;
		}
		elsif (ref $handler eq 'CODE') {
		    $self->DEBUG($self->ID, "about to call code $handler\n")
			if $DEBUG;
		    
		    eval {
			$result = &$handler($parser, \$dir);
			return $self->error('handler code returned undef')
			    unless defined $result;
		    };
		    return $self->error($@) if $@;
		}
		else {
		    $self->throw("invalid handler directive: $handler");
		}
		
		push(@{ $self->{ CONTENT } }, $result);
	    }
	    else {
		# didn't match a directive handler so reconstruct original 
		# directive and ask parser to parse as a statement, e.g. 
		# an expression, variable or assignment list

		$self->DEBUG("no handler for $keyword, \$parser->parse_statement('$dir')\n")
		    if $DEBUG;

		$result = $parser->parse_statement(\$dir);
		
		return $self->error($parser->error())
		    unless defined $result;

		push(@{ $self->{ CONTENT } }, $result);
	    }
	}    
	else {
	    $self->DEBUG("parsing statement by default\n")
		if $DEBUG;

	    $result = $parser->parse_statement(\$dir);

	    return $self->error($parser->error())
		unless defined $result;

	    push(@{ $self->{ CONTENT } }, $result);
	}

	if ($dir =~ / \G \s* ([^;]+) /gcx) {
	    $parser->parse_warning("ignored trailing cruft in directive: $1");
	}

	if ($dir =~ / \G \s* ; \s* (.+) /gx) {
	    $dir = $1;
	}
	else {
	    last;
	}
    }

    return 1;
}







#========================================================================
# debug/housekeeping methods
#========================================================================

#------------------------------------------------------------------------
# ID
#
# Return a useful string describing $self for debugging purposes.
#------------------------------------------------------------------------

sub ID {
    my $self = shift;
    my $ref = "$self";
    $ref = $1
	if $ref =~ /\((.*?)\)/;
    return "parser_handler<$ref>";
}


1;

__END__

