#============================================================= -*-perl-*-
#
# Template::Service
#
# DESCRIPTION
#   Object class to represent a runtime template processing service.
#
# AUTHOR
#   Andy Wardley   <abw@kfs.org>
#
# COPYRIGHT
#   Copyright (C) 2001 Andy Wardley.  All Rights Reserved.
#
#   This module is free software; you can redistribute it and/or
#   modify it under the same terms as Perl itself.
#
# REVISION
#   $Id: Service.pm,v 1.2 2001/06/19 16:15:25 abw Exp $
#
#========================================================================

package Template::Service;

use strict;
use Template::Base;
use vars qw( $VERSION $DEBUG $ERROR $ETYPE );
use base qw( Template::Base );

$VERSION = sprintf("%d.%02d", q$Revision: 1.2 $ =~ /(\d+)\.(\d+)/);
$DEBUG   = 0 unless defined $DEBUG;
$ERROR   = '';
$ETYPE   = 'service';

sub init {
    my ($self, $config) = @_;

    my $contexts = $self->{ _CONTEXTS } = $config->{ CONTEXTS } || { };

    if ($DEBUG) {
	$self->DEBUG("contexts: {\n    ", 
		     join("\n    ", 
			  map { "$_ => $contexts->{$_}" }
			  keys %$contexts),
		     "\n}\n");
    }

    return $self;
}


sub context {
    my ($self, $name, @args) = @_;
    my ($context, $current);
    my $contexts = $self->{ _CONTEXTS };

#    return $chash->{ _CURRENT } || $chash->{ _DEFAULT }
#	unless $name;


    if ($name) {
	# look for new context to switch to
	$context = $contexts->{ $name }
	    ||= $self->new_context($name, @args)
		|| return;
    }
    elsif ($context = $contexts->{ _CURRENT }) {
	# current context is already engaged
	return $context;
    }
    else {
	# get default context to engage
	$context = $self->new_context(undef, @args)
	    || return;
    }
	
    # disengage any current context
    if ($current = $contexts->{ _CURRENT }) {
	$self->DEBUG("$self disengaging ", $current->ID(), "\n") if $DEBUG;
	    
	$current->disengage($self)
	    || return $self->error($current->error());
    }

    # engage new context
    $context->engage($self, @args)
	|| return $self->error($context->error());

    $contexts->{ _CURRENT } = $context;

    $self->DEBUG("$self engaged ", $context->ID(), "\n") if $DEBUG; 

    return $context;
}


sub new_context {
    my ($self, $name, @args) = @_;
    my $contexts = $self->{ _CONTEXTS };
    my $context;

    if ($name) {
	# currently fails all the time (only gets called if
	# context() method can't already find context in 
	# _CONTEXTS hash)
	$context = $contexts->{ $name }
	    || return $self->error("$name: cannot create context (not implemented)");
    }
    else {
	# look for 'default' entry which can be a direct reference 
	# to a context or an indirect reference by name to another 
	# context
	$name = $contexts->{ default };
	return $self->error('no default context defined')
	    unless defined $name;
	$context = ref $name ? $name : $contexts->{ $name };
	return $self->error("$name: no such context to default to")
	    unless $context;
    }
    return $context;
}


sub start {
    my ($self, @args) = @_;
    $self->DEBUG("service start\n") if $DEBUG;
    return 1;
}

sub stop {
    my ($self, @args) = @_;
    $self->DEBUG("service stop\n") if $DEBUG;
    return 1;
}



sub DESTROY {
    my $self = shift;
    my $contexts = $self->{ _CONTEXTS };
    my $current  = $contexts->{ _CURRENT };

    # disengage any current context
    if ($current) {
	$self->DEBUG("$self DESTROY disengaging ", $current->ID(), "\n") if $DEBUG;
	$current->disengage($self);
    }

    return 1;
}


sub throw {
    my $self = shift;
    $self->SUPER::throw('Service', @_)
	unless ref $_[0];
    $self->SUPER::throw(@_);
}

1;


__END__

Service defines _CONTEXT hash which maps names to Context references.
Context names leading '_' are for internal use:  '_DEFAULT' specifies
default context if not explicitly named, '_CURRENT' specifies current
context in scope.
