#============================================================= -*-Perl-*-
#
# Template::Stash::v3
#
# DESCRIPTION
#   Prototype for a new stash which avoids multiple method calls by
#   iterating directly through the array of variables nodes FSM like.
#
# AUTHOR
#   Andy Wardley   <abw@kfs.org>
#
# COPYRIGHT
#   Copyright (C) 1996-2001 Andy Wardley.  All Rights Reserved.
#   Copyright (C) 1998-2001 Canon Research Centre Europe Ltd.
#
#   This module is free software; you can redistribute it and/or
#   modify it under the same terms as Perl itself.
#
# REVISION
#   $Id: Stash.pm,v 1.1 2001/06/19 16:15:25 abw Exp $
#
#============================================================================

package Template::Stash::v3;

require 5.005;

use strict;
use vars qw( $VERSION $DEBUG $VMETHS );

$VERSION = sprintf("%d.%02d", q$Revision: 1.1 $ =~ /(\d+)\.(\d+)/);
$VMETHS = {
    REF     => \&vmeth_ref,
    ITEM    => \&vmeth_item,
    LIST    => \&vmeth_list,
    HASH    => \&vmeth_hash,
    LENGTH  => \&vmeth_length,
    SIZE    => \&vmeth_size,
    MAX     => \&vmeth_max,
    FIRST   => \&vmeth_first,
    LAST    => \&vmeth_last,
#   sort/nsort/reverse/join
#   pop/push/shift/unshift
#   keys/values/each/import
#   search/replace/grep

};

use constant UNDEF   => "UNDEF\n";
use constant ARRAY   => "ARRAY\n";
use constant PRIVATE => "PRIVATE\n";
use constant UNKNOWN => "UNKNOWN\n";




#========================================================================
#                      -----  CLASS METHODS -----
#========================================================================

#------------------------------------------------------------------------
# new(\%params)
#
# Constructor method which creates a new Template::Stash object.
# An optional hash reference may be passed containing variable 
# definitions that will be used to initialise the stash.
#
# Returns a reference to a newly created Template::Stash.
#------------------------------------------------------------------------

sub new {
    my $class  = shift;
    my $params = ref $_[0] eq 'HASH' ? shift(@_) : { @_ };

    my $self   = {
	global  => { },
	%$params,
	'_PARENT' => undef,
    };

    bless $self, $class;
}


#========================================================================
#                   -----  PUBLIC OBJECT METHODS -----
#========================================================================

#------------------------------------------------------------------------
# clone(\%params)
#
# Creates a copy of the current stash object to effect localisation 
# of variables.  The new stash is blessed into the same class as the 
# parent (which may be a derived class) and has a '_PARENT' member added
# which contains a reference to the parent stash that created it
# ($self).  This member is used in a successive declone() method call to
# return the reference to the parent.
# 
# A parameter may be provided which should reference a hash of 
# variable/values which should be defined in the new stash.  The 
# update() method is called to define these new variables in the cloned
# stash.
#
# Returns a reference to a cloned Template::Stash.
#------------------------------------------------------------------------

sub clone {
    my ($self, $params) = @_;
    $params ||= { };

    # look out for magical 'import' argument which imports another hash
    my $import = $params->{ import };
    if (defined $import && UNIVERSAL::isa($import, 'HASH')) {
	delete $params->{ import };
    }
    else {
	undef $import;
    }

    my $clone = bless { 
	%$self,			# copy all parent members
	%$params,		# copy all new data
        '_PARENT' => $self,     # link to parent
    }, ref $self;
    
    # perform hash import if defined
#    &{ $HASH_OPS->{ import }}($clone, $import)
#	if defined $import;

    return $clone;
}

	
#------------------------------------------------------------------------
# declone($export) 
#
# Returns a reference to the PARENT stash.  When called in the following
# manner:
#    $stash = $stash->declone();
# the reference count on the current stash will drop to 0 and be "freed"
# and the caller will be left with a reference to the parent.  This 
# contains the state of the stash before it was cloned.  
#------------------------------------------------------------------------

sub declone {
    my $self = shift;
    $self->{ _PARENT } || $self;
}


#========================================================================

*get = \&operate;

sub set {
    my ($self, $ident, $value, $default) = @_;
    $ident = [ $ident, 0 ] unless ref $ident eq 'ARRAY';
    unshift(@$ident, 'LVALUE', 1);
    push(@$ident, 'ASSIGN', $value);
    $self->operate($ident);
}



sub operate {
    my ($self, $nodes, $optargs) = @_;
    my $root = $self;
    my ($rootref, $node, $args, $next, $action, $value, $last, $method);
    my ($prevroot, $prevnode, $prevargs);
    my ($default, $lvalue) = (0) x 2;

    $nodes = [ $nodes, $optargs || 0 ] unless ref $nodes eq 'ARRAY';

    my $i = 1;
    eval {
	while (@$nodes) {
	    ($node, $args) = splice(@$nodes, 0, 2);

#	    printf(STDERR "STATE %03d [$node] [%s]\n", $i++, 
#		   ref $args eq 'ARRAY' ? join(', ', @$args) : "'$args'");

	    die UNDEF unless defined($root) 
		and defined($node);

	    die PRIVATE if $node =~ /^[\._]/;

	    $rootref = ref $root;
	    $next = @$nodes ? $nodes->[0] : '';

	    # lookahead to detect a 'CODEREF' or 'ASSIGN' node in the 
	    # following node.  These cause us to delay evaluation of this
	    # node and instead construct a stash closure (array ref of 
	    # [ $root, $node, \@args ] which is then set to the $root 
	    # item for the next node (e.g. the actual CODEREF or ASSIGN)
	    # to unpack and manipulate as required.

#	     if ($next =~ /CODEREF|ASSIGN|DEFAULT/) {
#		 my ($copyroot, $copynode) = ($root, $node);
##		my $copyargs = ref $args eq 'ARRAY' ? [@$args] : $args;
#		 my @copyargs = ($args ? @$args : ());
#		 $root = [ $copyroot, $copynode, \@copyargs ];
##		$root = [ $root, $node, $args ];
#		 $self->DEBUG("next [$next] => [ $copyroot, $copynode, [@copyargs] ]\n")
#		     if $DEBUG;
#		 next;
#	     }
#
	    # jump table to preform various internal operations.
	    if ($node =~ /^ROOT|CLOSE|LVALUE|DEFAULT|ASSIGN|CODE(REF)?|METHOD$/) {
		goto $node;
	    }
	    # else drop through to DOTOP...

	  DOTOP:
	    if ($action = $VMETHS->{ $node }) {
		$root = &$action($self, $root, $node, $args, $nodes);
	    }
	    elsif ($rootref eq __PACKAGE__ || $rootref eq 'HASH') {
		if (defined($value = $root->{ $node })) {
		    if (ref $value eq 'CODE') {
			unshift(@$nodes, 'CODE', 0, $value, 
				$args ? [@$args] : 0);
		    }
		    else {
			$root = $value;
		    }
		    next;
		}
		elsif ($lvalue) {
		    # we create an intermediate hash if this is an lvalue
		    $root = ($root->{ $node } = { });
		    next;
		}
		die UNDEF;
	    }
	    elsif ($rootref eq 'ARRAY') {
		if ($node =~ /^\d+$/) {
		    $value = $root->[$node];
		    if (ref $value eq 'CODE') {
			unshift(@$nodes, 'CODE', 0, $value, 
				$args ? [@$args] : 0);
		    }
		    else {
			$root = $value;
		    }
		    next;
		}
		else {
		    die ARRAY;
		}
	    }
	    elsif (ref($root) && UNIVERSAL::can($root, 'can')) {
		# $root is an object, i.e. a reference that can('can')
		unshift(@$nodes, 'METHOD', $root, $node, 
			$args ? [@$args] : 0);
		next;
	    }
	    else {
		die UNKNOWN . "($root, $node)\n";
	    }
	    next;

	  ROOT:
	    $self->DEBUG("setting \$root to $args\n")
		if $DEBUG;
	    $root = $args;
	    next;

	  CLOSE:
	    my ($copyroot, $copynode) = ($root, $node);
	    my @copyargs = ($args ? @$args : ());
	    $root = [ $copyroot, $copynode, \@copyargs ];
	    $self->DEBUG("CLOSE => [ $copyroot, $copynode, [@copyargs] ]\n")
		if $DEBUG;
	    next;

	  LVALUE:
	    $self->DEBUG("setting \$lvalue to $args\n")
		if $DEBUG;
	    $lvalue = $args;
	    next;

	  DEFAULT:
	    $default++;
	    $self->DEBUG("incrementing \$default flag to $default\n")
		if $DEBUG;
	    # drop-through to ASSIGN

	  ASSIGN:
	    ($prevroot, $prevnode, $prevargs) = @$root;
	    $prevargs ||= [ ];
	    push(@$prevargs, $args);
	    $value = $prevargs->[0];

	    $self->DEBUG("ASSIGN $prevroot.$prevnode(@$prevargs, [$args]) => $value\n")
		if $DEBUG;

	    $rootref = ref $prevroot;
	    if ($rootref eq 'HASH' || $rootref eq __PACKAGE__) {
		# TODO: IMPORT!
		$root = ($default && $prevroot->{ $prevnode }
			 || ($prevroot->{ $prevnode } = $value));
	    }
	    elsif ($rootref eq 'ARRAY' && $prevnode =~ /^\d+$/) {
		$root = ($default && $prevroot->[ $prevnode ]
			 || ($prevroot->[ $prevnode ] = $value));
	    }
	    elsif (UNIVERSAL::can($prevroot, ($method = $prevnode))
	        || UNIVERSAL::can($prevroot, 'TT_CAN') 
		&& ($method = $prevroot->TT_CAN($method))) {
		$root = $prevroot->$method(@$prevargs);
	    }
	    else {
		$root = undef;
		die &UNKNOWN . "($prevroot, $prevnode)\n";
	    }
	    next;
	    
	  CODE:
#	    ($prevroot, $prevnode, $prevargs) = @$root;
	    my ($code, $codeargs) = splice(@$nodes, 0, 2);
	    $codeargs ||= [];

	    $self->DEBUG("CODE => $code(@$codeargs)\n")
		if $DEBUG;

	    # if the next node is 'REF' then result is 'CODE'
	    if (@$nodes && $nodes->[0] eq 'REF') {
		$root = $code;
	    }
	    else {
		$root = &$code(@$codeargs);
	    }
	    next;

	  CODEREF:
	    my ($closeroot, $closenode, $closeargs) = @$root;
	    $closeargs ||= [];
	    push(@$closeargs, @$args) if $args;
	    
	    $self->DEBUG("CODEREF => [$closeroot, $closenode, [@$closeargs]]\n")
		if $DEBUG;

	    $root = sub {
		my @args = (@$closeargs, @_);
		return $self->operate(['ROOT', $closeroot, 
				       $closenode, \@args]);
	    };
	    next;
	    
	  METHOD:
	    my ($method, $methargs) = splice(@$nodes, 0, 2);
	    $methargs ||= [];
	    $self->DEBUG("METHOD => $root->$method(@$methargs)\n")
		if $DEBUG;
	    $root = $root->$method(@$methargs);
	    next;
	}
	continue {
	    $last = $node if @$nodes;
	}
    };
    if ($@ eq UNDEF || ! defined $root) {
	$node = '<undef>' unless defined $node;
	$args = ref $args eq 'ARRAY' ? ( '(' . join(', ', @$args) . ')' ) : "($args)";
	$self->DEBUG("undefined value: ", $last ? "$last.$node$args" : "$node$args", "\n");
	return '';
    }
    elsif ($@ eq PRIVATE) {
	$self->DEBUG("attempt to access private variable: $last.$node\n");
	return '';
    }
    elsif ($@ eq ARRAY) {
	$self->DEBUG("cannot access non-numerical array item: $last.$node\n");
	return '';
    }
    elsif ($@ =~ s/^UNKNOWN\n//) {
	chomp $@;
	$self->DEBUG("don't known how to access: $last.$node $@\n");
	return '';
    }
    elsif ($@) {
	$self->DEBUG("error: $@\n");
	return '';
    }

    return $root;
}


#========================================================================
#                          --- VIRTUAL METHODS ---
#========================================================================

#------------------------------------------------------------------------
# vmeth_ref($root, 'REF', \@args, \@nodes)			      REF
#
# Returns a string indicating the type of data object that the $root
# node references.
#
# item.REF => ''
# list.REF => 'LIST'	    # NOTE: not ARRAY
# hash.REF => 'HASH'
# code.REF => 'CODE'
# obj.REF(args) => $obj->TT_REF(@$args) if $obj->can('TT_REF')
# obj.REF(args) => ref $obj if ! $obj->can('TT_REF')
# ???.REF => ref ???
#------------------------------------------------------------------------

sub vmeth_ref {
    my ($self, $root, $node, $args, $nodes) = @_;
    local $" = ', ';
    $args ||= [];

    $self->DEBUG("vmeth_ref($root, $node, @$args)\n")
	if $DEBUG;

    my $ref = ref $root;
    return '' unless $ref;

    # object can implement TT_REF() method
    return $root->TT_REF(@$args)
	if UNIVERSAL::can($root, 'TT_REF');

    return 'LIST' if $ref eq 'ARRAY';
    return $ref;
}


#------------------------------------------------------------------------
# vmeth_item($root, 'ITEM', \@args, \@nodes)			     ITEM
#
# item.ITEM => $item
# hash.ITEM => $hash
# list.ITEM => $list
# hash.ITEM(key) => $hash->{ key }
# list.ITEM(n)   => $list->[n]
# obj.ITEM(args) => $obj->TT_ITEM(@$args) if $obj->can('TT_ITEM')
# obj.ITEM(args) => $obj if ! $obj->can('TT_ITEM')
# ???.ITEM => ???
#------------------------------------------------------------------------

sub vmeth_item {
    my ($self, $root, $node, $args, $nodes) = @_;
    $args ||= [ ];
    $self->DEBUG("ITEM(root=$root, node=$node, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.ITEM returns self
    return $root unless ref $root;

    # object can implement TT_ITEM() method
    if (UNIVERSAL::can($root, 'TT_ITEM')) {
	return $root->TT_ITEM(@$args);
    }
    elsif (UNIVERSAL::isa($root, 'HASH')) {
	return @$args ? $root->{ $args->[0] || '' } : $root;
    }
    elsif (UNIVERSAL::isa($root, 'ARRAY')) {
	return @$args ? $root->[ $args->[0] || 0 ] : $root;
    }
    else {
#	die UNKNOWN;
	return $root;	    # code ref?
    }
}


#------------------------------------------------------------------------
# vmeth_list($root, 'LIST', \@args, \@nodes)			     LIST
#
# item.LIST => [ $item ]
# list.LIST => $list
# hash.LIST => [ { key=>..., value=>... }, { key=>..., value=>... }, ... ]
# hash.LIST('keys')   => [ keys %$hash ]
# hash.LIST('values') => [ values %$hash ]
# hash.LIST('each')   => [ %$hash ]
# obj.LIST(args) => $obj->TT_LIST(@$args) if $obj->can('TT_LIST')
# obj.LIST(args) => [ $obj ] if ! $obj->can('TT_LIST')
# ???.LIST => [ ??? ]
#------------------------------------------------------------------------

sub vmeth_list {
    my ($self, $root, $node, $args, $nodes) = @_;
    $args ||= [];
    $self->DEBUG("LIST(root=$root, node=$node, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.LIST returns [ item ]
    return [ $root ] unless ref $root;

    # object can implement TT_LIST() method
    if (UNIVERSAL::can($root, 'TT_LIST')) {
	return $root->TT_LIST(@$args);
    }
    elsif (UNIVERSAL::isa($root, 'ARRAY')) {
	return $root;
    }
    # don't inspect objects
    elsif (ref $root eq 'HASH') {
	my $what = $args->[0] || '';
	return ($what eq 'keys')   ? [   keys %$root ]
	     : ($what eq 'values') ? [ values %$root ]
	     : ($what eq 'each')   ? [ %$root ]
	     : [ map { { key => $_ , value => $root->{ $_ } } }
		 sort keys %$root ];
    }
    else {
#	die UNKNOWN;
	return [ $root ];	    # code ref?
    }

    die 'not reached';
}


#------------------------------------------------------------------------
# vmeth_hash($root, 'HASH', \@args, \@nodes)			     HASH
#
# item.HASH => { item => $item }
# list.HASH => { @$list }
# hash.HASH => $hash
# obj.HASH(args) => $obj->TT_HASH(@$args) if $obj->can('TT_HASH')
# obj.HASH(args) => [ $obj ] if ! $obj->can('TT_HASH')
# ???.HASH => { item => ??? }
#------------------------------------------------------------------------

sub vmeth_hash {
    my ($self, $root, $node, $args, $nodes) = @_;
    $args ||= [];
    $self->DEBUG("HASH(root=$root, node=$node, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.HASH returns { item => item }
    return { item => $root } unless ref $root;

    # object can implement TT_HASH() method
    if (UNIVERSAL::can($root, 'TT_HASH')) {
	return $root->TT_HASH(@$args);
    }
    elsif (ref $root eq 'HASH') {
	return $root;
    }
    # don't inspect objects
    elsif (ref $root eq 'ARRAY') {
	my @items = @$root;
	push(@items, '') unless $#items % 2;
	return { @items };
    }
    else {
#	die UNKNOWN;
	return { item => $root };	    # code ref?
    }

    die 'not reached';
}


#------------------------------------------------------------------------
# vmeth_length($root, 'LENGTH', \@args, \@nodes)	           LENGTH
#
# item.LENGTH => length $item
# list.LENGTH => scalar @$list
# hash.LENGTH => scalar keys %$hash
# obj.LENGTH(args) => $obj->TT_LENGTH(@$args) if $obj->can('TT_LENGTH')
# obj.LENGTH(args) => 1 if ! $obj->can('TT_LENGTH')
# ???.LENGTH => 1
#------------------------------------------------------------------------

sub vmeth_length {
    my ($self, $root, $node, $args, $nodes) = @_;
    $args ||= [];
    $self->DEBUG("LENGTH(root=$root, node=$node, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.LENGTH returns length of item
    return length $root unless ref $root;

    if (UNIVERSAL::can($root, 'TT_LENGTH')) {
	return $root->TT_LENGTH(@$args);
    }
    elsif (ref $root eq 'HASH') {
	return scalar keys %$root;
    }
    elsif (ref $root eq 'ARRAY') {
	return scalar @$root;
    }
    else {
	return 1;
    }

    die 'not reached';
}


#------------------------------------------------------------------------
# vmeth_size($root, 'SIZE', \@args, \@nodes)	                     SIZE
#
# item.SIZE => 1
# list.SIZE => scalar @$list
# hash.SIZE => scalar keys %$hash
# obj.SIZE(args) => $obj->TT_SIZE(@$args) if $obj->can('TT_SIZE')
# obj.SIZE(args) => 1 if ! $obj->can('TT_SIZE')
# ???.SIZE => 1
#------------------------------------------------------------------------

sub vmeth_size {
    my ($self, $root, $node, $args, $nodes) = @_;
    $args ||= [];
    $self->DEBUG("SIZE(root=$root, node=$node, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.SIZE returns 1 for a single item
    return 1 unless ref $root;

    if (UNIVERSAL::can($root, 'TT_SIZE')) {
	return $root->TT_SIZE(@$args);
    }
    elsif (ref $root eq 'HASH') {
	return scalar keys %$root;
    }
    elsif (ref $root eq 'ARRAY') {
	return scalar @$root;
    }
    else {
	return 1;
    }

    die 'not reached';
}


#------------------------------------------------------------------------
# vmeth_max($root, 'MAX', \@args, \@nodes)	                      MAX
#
# item.MAX => 1
# list.MAX => $#$list
# hash.MAX => scalar keys %$hash - 1
# obj.MAX(args) => $obj->TT_MAX(@$args) if $obj->can('TT_MAX')
# obj.MAX(args) => 0 if ! $obj->can('TT_MAX')
# ???.MAX => 0
#------------------------------------------------------------------------

sub vmeth_max {
    my ($self, $root, $node, $args, $nodes) = @_;
    $args ||= [];
    $self->DEBUG("MAX(root=$root, node=$node, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.MAX returns 1 for a single item
    return 0 unless ref $root;

    if (UNIVERSAL::can($root, 'TT_MAX')) {
	return $root->TT_MAX(@$args);
    }
    elsif (ref $root eq 'HASH') {
	return (scalar keys %$root) - 1;
    }
    elsif (ref $root eq 'ARRAY') {
	return $#$root;
    }
    else {
	return 0;
    }

    die 'not reached';
}


#------------------------------------------------------------------------
# vmeth_first($root, 'FIRST', \@args, \@nodes)	                    FIRST
#
# item.FIRST => $item
# list.FIRST => $list->[0]
# hash.FIRST => hash.LIST.FIRST
# obj.FIRST(args) => $obj->TT_FIRST(@$args) if $obj->can('TT_FIRST')
# obj.FIRST(args) => $obj if ! $obj->can('TT_FIRST')
# ???.FIRST => ???
#------------------------------------------------------------------------

sub vmeth_first {
    my ($self, $root, $node, $args, $nodes) = @_;
    $args ||= [];
    $self->DEBUG("FIRST(root=$root, node=$node, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.FIRST returns single item intact
    return $root unless ref $root;

    if (UNIVERSAL::can($root, 'TT_FIRST')) {
	return $root->TT_FIRST(@$args);
    }
    elsif (ref $root ne 'ARRAY') {
	my $prev = $root;
	$root = $self->vmeth_list($root);
	$self->DEBUG("cast $prev to $root\n")
	    if $DEBUG;
    }

    return $root->[0];
}


#------------------------------------------------------------------------
# vmeth_last($root, 'LAST', \@args, \@nodes)	                     LAST
#
# item.LAST => $item
# list.LAST => $list->[-1]
# hash.LAST => hash.LIST.LAST
# obj.LAST(args) => $obj->TT_LAST(@$args) if $obj->can('TT_LAST')
# obj.LAST(args) => $obj if ! $obj->can('TT_LAST')
# ???.LAST => ???
#------------------------------------------------------------------------

sub vmeth_last {
    my ($self, $root, $node, $args, $nodes) = @_;
    $args ||= [];
    $self->DEBUG("LAST(root=$root, node=$node, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.LAST returns single item intact
    return $root unless ref $root;

    if (UNIVERSAL::can($root, 'TT_LAST')) {
	return $root->TT_LAST(@$args);
    }
    elsif (ref $root ne 'ARRAY') {
	my $prev = $root;
	$root = $self->vmeth_list($root);
	$self->DEBUG("cast $prev to $root\n")
	    if $DEBUG;
    }

    return $root->[-1];
}

#========================================================================

#------------------------------------------------------------------------
# update(\%params)
#
# Update multiple variables en masse.  No magic is performed.  Simple
# variable names only.
#------------------------------------------------------------------------

sub update {
    my ($self, $params) = @_;

    # look out for magical 'import' argument to import another hash
    my $import = $params->{ import };
    if (defined $import && UNIVERSAL::isa($import, 'HASH')) {
	@$self{ keys %$import } = values %$import;
	delete $params->{ import };
    }

    @$self{ keys %$params } = values %$params;
}




#------------------------------------------------------------------------
# _dump()
#
# Debug method which returns a string representing the internal state
# of the object.  The method calls itself recursively to dump sub-hashes.
#------------------------------------------------------------------------

sub _dump {
    my $self   = shift;
    my $indent = shift || 1;
    my $buffer = '    ';
    my $pad    = $buffer x $indent;
    my $text   = '';
    local $" = ', ';

    my ($key, $value);


    return $text . "...excessive recursion, terminating\n"
	if $indent > 32;

    foreach $key (keys %$self) {

	$value = $self->{ $key };
	$value = '<undef>' unless defined $value;

	if (ref($value) eq 'ARRAY') {
	    $value = "$value [@$value]";
	}
	$text .= sprintf("$pad%-8s => $value\n", $key);
	next if $key =~ /^\./;
	if (UNIVERSAL::isa($value, 'HASH')) {
	    $text .= _dump($value, $indent + 1);
	}
    }
    $text;
}

sub DEBUG {
    my $self = shift;
    print STDERR @_;
}


1;

__END__


#------------------------------------------------------------------------
# IMPORTANT NOTE
#   This documentation is generated automatically from source
#   templates.  Any changes you make here may be lost.
# 
#   The 'docsrc' documentation source bundle is available for download
#   from http://www.template-toolkit.org/docs.html and contains all
#   the source templates, XML files, scripts, etc., from which the
#   documentation for the Template Toolkit is built.
#------------------------------------------------------------------------

=head1 NAME

Template::Stash - Magical storage for template variables

=head1 SYNOPSIS

    use Template::Stash;

    my $stash = Template::Stash->new(\%vars);

    # get variable values
    $value = $stash->get($variable);
    $value = $stash->get(\@compound);

    # set variable value
    $stash->set($variable, $value);
    $stash->set(\@compound, $value);

    # default variable value
    $stash->set($variable, $value, 1);
    $stash->set(\@compound, $value, 1);

    # set variable values en masse
    $stash->update(\%new_vars)

    # methods for (de-)localising variables
    $stash = $stash->clone(\%new_vars);
    $stash = $stash->declone();

=head1 DESCRIPTION

The Template::Stash module defines an object class which is used to store
variable values for the runtime use of the template processor.  Variable
values are stored internally in a hash reference (which itself is blessed 
to create the object) and are accessible via the get() and set() methods.

Variables may reference hash arrays, lists, subroutines and objects
as well as simple values.  The stash automatically performs the right
magic when dealing with variables, calling code or object methods,
indexing into lists, hashes, etc.

The stash has clone() and declone() methods which are used by the
template processor to make temporary copies of the stash for
localising changes made to variables.

=head1 PUBLIC METHODS

=head2 new(\%params)

The new() constructor method creates and returns a reference to a new
Template::Stash object.  

    my $stash = Template::Stash->new();

A hash reference may be passed to provide variables and values which
should be used to initialise the stash.

    my $stash = Template::Stash->new({ var1 => 'value1', 
				       var2 => 'value2' });

=head2 get($variable)

The get() method retrieves the variable named by the first parameter.

    $value = $stash->get('var1');

Dotted compound variables can be retrieved by specifying the variable
elements by reference to a list.  Each node in the variable occupies
two entries in the list.  The first gives the name of the variable
element, the second is a reference to a list of arguments for that 
element, or 0 if none.

    [% foo.bar(10).baz(20) %]

    $stash->get([ 'foo', 0, 'bar', [ 10 ], 'baz', [ 20 ]);

=head2 set($variable, $value, $default)

The set() method sets the variable name in the first parameter to the 
value specified in the second.

    $stash->set('var1', 'value1');

Dotted compound variables may be specified as per get() above.

    [% foo.bar = 30 %]

    $stash->set([ 'foo', 0, 'bar', 0 ], 30);

The magical variable 'IMPORT' can be specified whose corresponding
value should be a hash reference.  The contents of the hash array are
copied (i.e. imported) into the current namespace.

    # foo.bar = baz, foo.wiz = waz
    $stash->set('foo', { 'bar' => 'baz', 'wiz' => 'waz' });

    # import 'foo' into main namespace: foo = baz, wiz = waz
    $stash->set('IMPORT', $stash->get('foo'));

=head2 clone(\%params)

The clone() method creates and returns a new Template::Stash object which
represents a localised copy of the parent stash.  Variables can be
freely updated in the cloned stash and when declone() is called, the
original stash is returned with all its members intact and in the
same state as they were before clone() was called.

For convenience, a hash of parameters may be passed into clone() which 
is used to update any simple variable (i.e. those that don't contain any 
namespace elements like 'foo' and 'bar' but not 'foo.bar') variables while 
cloning the stash.  For adding and updating complex variables, the set() 
method should be used after calling clone().  This will correctly resolve
and/or create any necessary namespace hashes.

A cloned stash maintains a reference to the stash that it was copied 
from in its '_PARENT' member.

=head2 declone()

The declone() method returns the '_PARENT' reference and can be used to
restore the state of a stash as described above.

=head1 AUTHOR

Andy Wardley E<lt>abw@kfs.orgE<gt>

L<http://www.andywardley.com/|http://www.andywardley.com/>

=head1 VERSION

2.17, distributed as part of the
Template Toolkit version 2.04, released on 12 June 2001.

=head1 COPYRIGHT

  Copyright (C) 1996-2001 Andy Wardley.  All Rights Reserved.
  Copyright (C) 1998-2001 Canon Research Centre Europe Ltd.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 SEE ALSO

L<Template|Template>, L<Template::Context|Template::Context>


