#============================================================= -*-Perl-*-
#
# Template::Stash::v3
#
# DESCRIPTION
#   Behold!  This is a totally new implementation of the stash for
#   version 3.  Credit is due to Craig Barratt for implementing
#   the lookahead functionality in the v2 stash (see the v2.03
#   Template::Stash::Context module) from which much of the 
#   following was inspired.
#
# AUTHOR
#   Andy Wardley   <abw@kfs.org>
#
# COPYRIGHT
#   Copyright (C) 1996-2001 Andy Wardley.  All Rights Reserved.
#   Copyright (C) 1998-2001 Canon Research Centre Europe Ltd.
#
#   This module is free software; you can redistribute it and/or
#   modify it under the same terms as Perl itself.
#
# TODO 
#   Support flags for any and all virtual methods that might require
#   sorting.  e.g. myhash.JOIN(sort => 'number', order => 'reverse')
#   this should get handled by vmeth_sort().
#
# REVISION
#   $Id: v3.pm,v 1.1 2001/06/19 16:15:25 abw Exp $
#
#============================================================================

package Template::Stash::v3;

require 5.005;

use strict;
use vars qw( $VERSION $DEBUG $WARNINGS
	     $VMETHS $HASH_EQUALS $HASH_JOINT $LIST_JOINT );

$VERSION  = sprintf("%d.%02d", q$Revision: 1.1 $ =~ /(\d+)\.(\d+)/);
$DEBUG    = 0 unless defined $DEBUG;
$WARNINGS = 0 unless defined $WARNINGS;

$VMETHS = {
    REF     => \&vmeth_ref,
    ITEM    => \&vmeth_item,
    LIST    => \&vmeth_list,
    HASH    => \&vmeth_hash,
    LENGTH  => \&vmeth_length,
    SIZE    => \&vmeth_size,
    MAX     => \&vmeth_max,
    FIRST   => \&vmeth_first,
    LAST    => \&vmeth_last,
    JOIN    => \&vmeth_join,
    SORT    => \&vmeth_sort,
    REVERSE => \&vmeth_reverse,
    REPEAT  => \&vmeth_repeat,
# TODO:  (need to think about hash/list conversion issues)
#   POP/PUSH/SHIFT/UNSHIFT
#   KEYS/VALUES/EACH/IMPORT
#   SEARCH/REPLACE/GREP
};

$HASH_EQUALS = ' = ' unless defined $HASH_EQUALS;
$HASH_JOINT  = ', '  unless defined $HASH_JOINT;
$LIST_JOINT  = ', '  unless defined $LIST_JOINT;

use constant UNDEF   => "UNDEF\n";
use constant ARRAY   => "ARRAY\n";
use constant PRIVATE => "PRIVATE\n";
use constant UNKNOWN => "UNKNOWN\n";




#========================================================================
#                      -----  CLASS METHODS -----
#========================================================================

#------------------------------------------------------------------------
# new(\%params)
#
# Constructor method which creates a new Template::Stash object.
# An optional hash reference may be passed containing variable 
# definitions that will be used to initialise the stash.
#
# Returns a reference to a newly created Template::Stash.
#------------------------------------------------------------------------

sub new {
    my $class  = shift;
    my $params = ref $_[0] eq 'HASH' ? shift(@_) : { @_ };

    my $self   = {
	global  => { },
	%$params,
	'_PARENT' => undef,
    };

    bless $self, $class;
}


#========================================================================
#                   -----  PUBLIC OBJECT METHODS -----
#========================================================================

#------------------------------------------------------------------------
# clone(\%params)
#
# Creates a copy of the current stash object to effect localisation 
# of variables.  The new stash is blessed into the same class as the 
# parent (which may be a derived class) and has a '_PARENT' member added
# which contains a reference to the parent stash that created it
# ($self).  This member is used in a successive declone() method call to
# return the reference to the parent.
# 
# A parameter may be provided which should reference a hash of 
# variable/values which should be defined in the new stash.  The 
# update() method is called to define these new variables in the cloned
# stash.
#
# Returns a reference to a cloned Template::Stash.
#------------------------------------------------------------------------

sub clone {
    my ($self, $params) = @_;
    $params ||= { };

    # look out for magical 'import' argument which imports another hash
    my $import = $params->{ import };
    if (defined $import && UNIVERSAL::isa($import, 'HASH')) {
	delete $params->{ import };
    }
    else {
	undef $import;
    }

    my $clone = bless { 
	%$self,			# copy all parent members
	%$params,		# copy all new data
        '_PARENT' => $self,     # link to parent
    }, ref $self;
    
    # perform hash import if defined
#    &{ $HASH_OPS->{ import }}($clone, $import)
#	if defined $import;

    return $clone;
}

	
#------------------------------------------------------------------------
# declone() 
#
# Returns a reference to the PARENT stash.  When called in the following
# manner:
#    $stash = $stash->declone();
# the reference count on the current stash will drop to 0 and be "freed"
# and the caller will be left with a reference to the parent.  This 
# contains the state of the stash before it was cloned.  
#------------------------------------------------------------------------

sub declone {
    my $self = shift;
    $self->{ _PARENT } || $self;
}


#========================================================================

*get = \&operate;

sub set {
    my ($self, $ident, $value, $default) = @_;
    $ident = [ $ident, 0 ] unless ref $ident eq 'ARRAY';
    unshift(@$ident, 'LVALUE', 1);
    push(@$ident, 'ASSIGN', $value);

    if ($DEBUG) {
	local $" = ', ';
	$self->DEBUG("set([@$ident] => $value\n");
    }

    $self->operate($ident);
}

sub pre_inc {
    my ($self, $ident, $inc) = @_;
    $inc = 1 unless defined $inc;

    # get() is destructive!
    my $copy = ref $ident eq 'ARRAY' ? [ @$ident ] : $ident;
    my $value = $self->get($ident);

    if ($DEBUG) {
	local $" = ', ';
	my $dbgid = ref $ident eq 'ARRAY' ? "[@$ident]" : $ident;
	$self->DEBUG("pre_inc($dbgid) $value => ", $value + $inc, "\n");
    }

    no warnings;
    $value += $inc;
    $self->set($copy, $value);
    return $value;
}

sub post_inc {
    my ($self, $ident, $inc) = @_;
    $inc = 1 unless defined $inc;

    # get() is destructive!
    my $copy = ref $ident eq 'ARRAY' ? [ @$ident ] : $ident;
    my $value = $self->get($ident);

    if ($DEBUG) {
	local $" = ', ';
	my $dbgid = ref $ident eq 'ARRAY' ? "[@$ident]" : $ident;
	$self->DEBUG("post_inc($dbgid) $value => ", $value + $inc, "\n");
    }

    no warnings;
    $self->set($copy, $value + $inc);
    return $value;
}


sub operate {
    my ($self, $nodes, $optargs) = @_;
    my $root = $self;
    my ($rootref, $node, $args, $next, $action, $value, $last, $method);
    my ($prevroot, $prevnode, $prevargs);
    my ($default, $lvalue) = (0) x 2;
    my (@result, $nitems);

    $nodes = [ $nodes, $optargs || 0 ] unless ref $nodes eq 'ARRAY';

    my $i = 1;
    eval {
	while (@$nodes) {
	    ($node, $args) = splice(@$nodes, 0, 2);

#	    printf(STDERR "STATE %03d [$node] [%s]\n", $i++, 
#		   ref $args eq 'ARRAY' ? join(', ', @$args) : "'$args'");

	    die UNDEF unless defined($root) 
		and defined($node);

	    die PRIVATE if $node =~ /^[\._]/;

	    $rootref = ref $root;
	    $next = @$nodes ? $nodes->[0] : '';

	    # lookahead to following node to detect a 'CODEREF', 'ASSIGN' 
	    # or 'REF'.  If found, then evaluation of the current node is
	    # delayed and instead $root is set to a stash closure (array ref
	    # of [ $root, $node, \@args ] for the next node (e.g. the actual 
	    # CODEREF, ASSIGN or DEFAULT) to unpack and manipulate as required.

	    if ($next =~ /CODEREF|ASSIGN|DEFAULT/) {
		my ($copyroot, $copynode) = ($root, $node);
		my @copyargs = ($args ? @$args : ());
		$root = [ $copyroot, $copynode, \@copyargs ];
		$self->DEBUG("lazy [$next] => [ $copyroot, $copynode, [@copyargs] ]\n")
		    if $DEBUG;
		next;
	    }

	    # jump table to preform various internal operations.
	    if ($node =~ /^ROOT|LVALUE|DEFAULT|ASSIGN|CODE(REF)?|METHOD$/) {
		# Bliss!  I'm using a computed goto and it's (IMHO) the
		# Right Thing to do.  Not often you get to do that. :-)
		goto $node;
	    }
	    # else drop through to DOTOP...

	  DOTOP:
	    if ($action = $VMETHS->{ $node }) {
		# can we delegate virtual method to an object?
		if (UNIVERSAL::can($root, ($method = "TT_$action"))) {
		    unshift(@$nodes, $method, $args);
		    goto METHOD;
		}
		# otherwise call inbuilt virtual method
		$root = &$action($self, $root, $args, $nodes);
	    }
	    elsif ($rootref eq __PACKAGE__ || $rootref eq 'HASH') {
		if (defined($value = $root->{ $node })) {
		    if (ref $value eq 'CODE') {
			unshift(@$nodes, 'CODE', 0, $value, 
				$args ? [@$args] : 0);
		    }
		    else {
			$root = $value;
		    }
		    next;
		}
		elsif ($lvalue) {
		    # we create an intermediate hash if this is an lvalue
		    $root = ($root->{ $node } = { });
		    next;
		}
		die UNDEF;
	    }
	    elsif ($rootref eq 'ARRAY') {
		if ($node =~ /^\d+$/) {
		    $value = $root->[$node];
		    if (ref $value eq 'CODE') {
			unshift(@$nodes, 'CODE', 0, $value, 
				$args ? [@$args] : 0);
		    }
		    else {
			$root = $value;
		    }
		    next;
		}
		else {
		    die ARRAY;
		}
	    }
	    elsif (ref($root) && UNIVERSAL::can($root, 'can')) {
		# $root is an object, i.e. a reference that can('can')
		unshift(@$nodes, 'METHOD', $root, $node, 
			$args ? [@$args] : 0);
		next;
	    }
	    else {
		die UNKNOWN . "($root, $node)\n";
	    }
	    next;

	  ROOT:
	    $self->DEBUG("setting \$root to $args\n")
		if $DEBUG;
	    $root = $args;
	    next;

	  LVALUE:
	    $self->DEBUG("setting \$lvalue to $args\n")
		if $DEBUG;
	    $lvalue = $args;
	    next;

	  DEFAULT:
	    $default++;
	    $self->DEBUG("incrementing \$default flag to $default\n")
		if $DEBUG;
	    # drop-through to ASSIGN

	  ASSIGN:
	    ($prevroot, $prevnode, $prevargs) = @$root;
	    $prevargs ||= [ ];
	    push(@$prevargs, $args);
	    $value = $prevargs->[0];

	    $self->DEBUG("ASSIGN $prevroot.$prevnode(@$prevargs, [$args]) => $value\n")
		if $DEBUG;

	    $rootref = ref $prevroot;
	    if ($rootref eq 'HASH' || $rootref eq __PACKAGE__) {
		# TODO: IMPORT!
		$root = ($default && $prevroot->{ $prevnode }
			 || ($prevroot->{ $prevnode } = $value));
	    }
	    elsif ($rootref eq 'ARRAY' && $prevnode =~ /^\d+$/) {
		$root = ($default && $prevroot->[ $prevnode ]
			 || ($prevroot->[ $prevnode ] = $value));
	    }
	    elsif (UNIVERSAL::can($prevroot, ($method = $prevnode))
		|| UNIVERSAL::can($prevroot, 'TT_CAN') 
	        && ($method = $prevroot->TT_CAN($method))) {
		$root = $prevroot->$method(@$prevargs);
	    }
	    else {
		$root = undef;
		die &UNKNOWN . "($prevroot, $prevnode)\n";
	    }
	    next;

	  CODEREF:
	    my ($closeroot, $closenode, $closeargs) = @$root;
	    $closeargs ||= [];
	    push(@$closeargs, @$args) if $args;
	    
	    $self->DEBUG("CODEREF => [$closeroot, $closenode, [@$closeargs]]\n")
		if $DEBUG;

	    $root = sub {
		my @args = (@$closeargs, @_);
		return $self->operate(['ROOT', $closeroot, 
				       $closenode, \@args]);
	    };
	    next;

	  CODE:
#	    ($prevroot, $prevnode, $prevargs) = @$root;
	    my ($code, $codeargs) = splice(@$nodes, 0, 2);
	    $codeargs ||= [];

	    $self->DEBUG("CODE => $code(@$codeargs)\n")
		if $DEBUG;

	    # if the next node is 'REF' then result is 'CODE'
	    if (@$nodes && (($next = $nodes->[0]) eq 'REF')) {
		$root = $code;
	    }
	    elsif ($next eq 'ITEM') {
		# call code in SCALAR context
		$root = scalar &$code(@$codeargs);
	    }
	    else {
		# call method in list context by default
		@result = &$code(@$codeargs);
		goto CHECK_RESULT;
	    }
	    next;

	  METHOD:
	    my ($method, $methargs) = splice(@$nodes, 0, 2);
	    $methargs ||= [];
	    $self->DEBUG("METHOD => $root->$method(@$methargs)\n")
		if $DEBUG;

	    # if the next node is 'REF' then result is 'CODE'
	    # we also splice off the REF op because it wouldn't
	    # otherwise know that this is a method 
	    if (@$nodes && (($next = $nodes->[0]) eq 'REF')) {
		splice(@$nodes, 0, 2);
		$root = 'CODE(' . ref($root) . ')';
	    }
	    elsif ($next eq 'ITEM') {
		# execute in SCALAR context
		$root = $root->$method(@$methargs);
	    }
	    else {
		# call method in list context by default
		@result = $root->$method(@$methargs);
		goto CHECK_RESULT;
	    }
	    next;

	  CHECK_RESULT:
	    $self->DEBUG("CHECK_RESULT [@result]\n")
		if $DEBUG;

	    if (($nitems = scalar @result) == 1) {
		$root = $result[0];
	    }
	    elsif ($nitems > 1) {
		$root = [ @result ];
	    }
	    else {
		$root = undef;
	    }
	    next;
	}
	continue {
	    $last = $node if @$nodes;
	}
    };
    if ($@ eq UNDEF || ! defined $root) {
	$node = '<undef>' unless defined $node;
	$args = ref $args eq 'ARRAY' ? ( '(' . join(', ', @$args) . ')' ) : "($args)";
	$self->WARNING("undefined value: ", $last ? "$last.$node$args" : "$node$args", "\n");
	return '';
    }
    elsif ($@ eq PRIVATE) {
	$self->WARNING("attempt to access private variable: $last.$node\n")
	    if $WARNINGS;
	return '';
    }
    elsif ($@ eq ARRAY) {
	$self->WARNING("cannot access non-numerical array item: $last.$node\n")
	    if $WARNINGS;
	return '';
    }
    elsif ($@ =~ s/^UNKNOWN\n//) {
	chomp $@;
	$self->WARNING("don't known how to access: $last.$node $@\n")
	    if $WARNINGS;
	return '';
    }
    elsif ($@) {
	$self->WARNING("error: $@\n")
	    if $WARNINGS;
	return '';
    }

    return $root;
}


#========================================================================
#                          --- VIRTUAL METHODS ---
#========================================================================

#------------------------------------------------------------------------
# vmeth_ref($root, \@args, \@nodes)			      REF
#
# Returns a string indicating the type of data object that the $root
# node references.
#
# item.REF => ''
# list.REF => 'LIST'	    # NOTE: not ARRAY
# hash.REF => 'HASH'
# code.REF => 'CODE'
# obj.REF(args) => $obj->TT_REF(@$args) if $obj->can('TT_REF')
# obj.REF(args) => ref $obj if ! $obj->can('TT_REF')
# ???.REF => ref ???
#------------------------------------------------------------------------

sub vmeth_ref {
    my ($self, $root, $args, $nodes) = @_;
    local $" = ', ';
    $args ||= [];

    $self->DEBUG("vmeth_ref($root, @$args)\n")
	if $DEBUG;

    my $ref = ref $root;
    return '' unless $ref;

    # object can implement TT_REF() method
    return $root->TT_REF(@$args)
	if UNIVERSAL::can($root, 'TT_REF');

    return 'LIST' if $ref eq 'ARRAY';
    return $ref;
}


#------------------------------------------------------------------------
# vmeth_item($root, \@args, \@nodes)			     ITEM
#
# item.ITEM => $item
# hash.ITEM => $hash
# list.ITEM => $list
# hash.ITEM(key) => $hash->{ key }
# list.ITEM(n)   => $list->[n]
# obj.ITEM(args) => $obj->TT_ITEM(@$args) if $obj->can('TT_ITEM')
# obj.ITEM(args) => $obj if ! $obj->can('TT_ITEM')
# ???.ITEM => ???
#------------------------------------------------------------------------

sub vmeth_item {
    my ($self, $root, $args, $nodes) = @_;
    $args ||= [ ];
    $self->DEBUG("ITEM(root=$root, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.ITEM returns self
    return $root unless ref $root;

    # object can implement TT_ITEM() method
    if (UNIVERSAL::can($root, 'TT_ITEM')) {
	return $root->TT_ITEM(@$args);
    }
    elsif (UNIVERSAL::isa($root, 'HASH')) {
	return @$args ? $root->{ $args->[0] || '' } : $root;
    }
    elsif (UNIVERSAL::isa($root, 'ARRAY')) {
	return @$args ? $root->[ $args->[0] || 0 ] : $root;
    }
    else {
#	die UNKNOWN;
	return $root;	    # code ref?
    }
}


#------------------------------------------------------------------------
# vmeth_list($root, \@args, \@nodes)			     LIST
#
# item.LIST => [ $item ]
# list.LIST => $list
# hash.LIST() => [ %$hash ]
# hash.LIST('keys')   => [ keys %$hash ]
# hash.LIST('values') => [ values %$hash ]
# hash.LIST('each')   => [ %$hash ]
# hash.LIST('hashrefs') => [ { key=>..., value=>... }, { key=>..., value=>... }, ... ]
# obj.LIST(args) => $obj->TT_LIST(@$args) if $obj->can('TT_LIST')
# obj.LIST(args) => [ $obj ] if ! $obj->can('TT_LIST')
# ???.LIST => [ ??? ]
#------------------------------------------------------------------------

sub vmeth_list {
    my ($self, $root, $args, $nodes) = @_;
    $args ||= [];
    $self->DEBUG("LIST(root=$root, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.LIST returns [ item ]
    return [ $root ] unless ref $root;

    # object can implement TT_LIST() method
    if (UNIVERSAL::can($root, 'TT_LIST')) {
	return $root->TT_LIST(@$args);
    }
    elsif (UNIVERSAL::isa($root, 'ARRAY')) {
	return $root;
    }
    # don't inspect objects
    elsif (ref $root eq 'HASH') {
	my $what = $args->[0] || '';
	return ($what eq 'keys')     ? [   keys %$root ]
	     : ($what eq 'values')   ? [ values %$root ]
	     : ($what eq 'each')     ? [        %$root ]
	     : ($what eq 'hashrefs') ? [ 
		    map {
			{ key => $_ , value => $root->{ $_ } }
		    } sort keys %$root ]
	     : [ %$root ];
    }
    else {
#	die UNKNOWN;
	return [ $root ];	    # code ref?
    }

    die 'not reached';
}


#------------------------------------------------------------------------
# vmeth_hash($root, \@args, \@nodes)			     HASH
#
# item.HASH => { item => $item }
# list.HASH => { @$list }
# hash.HASH => $hash
# obj.HASH(args) => $obj->TT_HASH(@$args) if $obj->can('TT_HASH')
# obj.HASH(args) => [ $obj ] if ! $obj->can('TT_HASH')
# ???.HASH => { item => ??? }
#------------------------------------------------------------------------

sub vmeth_hash {
    my ($self, $root, $args, $nodes) = @_;
    $args ||= [];
    $self->DEBUG("HASH(root=$root, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.HASH returns { item => item }
    return { item => $root } unless ref $root;

    # object can implement TT_HASH() method
    if (UNIVERSAL::can($root, 'TT_HASH')) {
	return $root->TT_HASH(@$args);
    }
    elsif (ref $root eq 'HASH') {
	return $root;
    }
    # don't inspect objects
    elsif (ref $root eq 'ARRAY') {
	my @items = @$root;
	push(@items, '') unless $#items % 2;
	return { @items };
    }
    else {
#	die UNKNOWN;
	return { item => $root };	    # code ref?
    }

    die 'not reached';
}


#------------------------------------------------------------------------
# vmeth_length($root, \@args, \@nodes)	           LENGTH
#
# item.LENGTH => length $item
# list.LENGTH => scalar @$list
# hash.LENGTH => scalar keys %$hash
# obj.LENGTH(args) => $obj->TT_LENGTH(@$args) if $obj->can('TT_LENGTH')
# obj.LENGTH(args) => 1 if ! $obj->can('TT_LENGTH')
# ???.LENGTH => 1
#------------------------------------------------------------------------

sub vmeth_length {
    my ($self, $root, $args, $nodes) = @_;
    $args ||= [];
    $self->DEBUG("LENGTH(root=$root, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.LENGTH returns length of item
    return length $root unless ref $root;

    if (UNIVERSAL::can($root, 'TT_LENGTH')) {
	return $root->TT_LENGTH(@$args);
    }
    elsif (ref $root eq 'HASH') {
	return scalar keys %$root;
    }
    elsif (ref $root eq 'ARRAY') {
	return scalar @$root;
    }
    else {
	return 1;
    }

    die 'not reached';
}


#------------------------------------------------------------------------
# vmeth_size($root, \@args, \@nodes)	                     SIZE
#
# item.SIZE => 1
# list.SIZE => scalar @$list
# hash.SIZE => scalar keys %$hash
# obj.SIZE(args) => $obj->TT_SIZE(@$args) if $obj->can('TT_SIZE')
# obj.SIZE(args) => 1 if ! $obj->can('TT_SIZE')
# ???.SIZE => 1
#------------------------------------------------------------------------

sub vmeth_size {
    my ($self, $root, $args, $nodes) = @_;
    $args ||= [];
    $self->DEBUG("SIZE(root=$root, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.SIZE returns 1 for a single item
    return 1 unless ref $root;

    if (UNIVERSAL::can($root, 'TT_SIZE')) {
	return $root->TT_SIZE(@$args);
    }
    elsif (ref $root eq 'HASH') {
	return scalar keys %$root;
    }
    elsif (ref $root eq 'ARRAY') {
	return scalar @$root;
    }
    else {
	return 1;
    }

    die 'not reached';
}


#------------------------------------------------------------------------
# vmeth_max($root, \@args, \@nodes)	                      MAX
#
# item.MAX => 1
# list.MAX => $#$list
# hash.MAX => scalar keys %$hash - 1
# obj.MAX(args) => $obj->TT_MAX(@$args) if $obj->can('TT_MAX')
# obj.MAX(args) => 0 if ! $obj->can('TT_MAX')
# ???.MAX => 0
#------------------------------------------------------------------------

sub vmeth_max {
    my ($self, $root, $args, $nodes) = @_;
    $args ||= [];
    $self->DEBUG("MAX(root=$root, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.MAX returns 1 for a single item
    return 0 unless ref $root;

    if (UNIVERSAL::can($root, 'TT_MAX')) {
	return $root->TT_MAX(@$args);
    }
    elsif (ref $root eq 'HASH') {
	return (scalar keys %$root) - 1;
    }
    elsif (ref $root eq 'ARRAY') {
	return $#$root;
    }
    else {
	return 0;
    }

    die 'not reached';
}


#------------------------------------------------------------------------
# vmeth_first($root, \@args, \@nodes)	                    FIRST
#
# item.FIRST => $item
# list.FIRST => $list->[0]
# hash.FIRST => hash.LIST.FIRST
# obj.FIRST(args) => $obj->TT_FIRST(@$args) if $obj->can('TT_FIRST')
# obj.FIRST(args) => $obj if ! $obj->can('TT_FIRST')
# ???.FIRST => ???
#------------------------------------------------------------------------

sub vmeth_first {
    my ($self, $root, $args, $nodes) = @_;
    $args ||= [];
    $self->DEBUG("FIRST(root=$root, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.FIRST returns single item intact
    return $root unless ref $root;

    return $root->TT_FIRST(@$args)
	if UNIVERSAL::can($root, 'TT_FIRST');

    $root = $self->vmeth_list($root)
	unless ref $root eq 'ARRAY';

    return $root->[0];
}


#------------------------------------------------------------------------
# vmeth_last($root, \@args, \@nodes)	                     LAST
#
# item.LAST => $item
# list.LAST => $list->[-1]
# hash.LAST => hash.LIST.LAST
# obj.LAST(args) => $obj->TT_LAST(@$args) if $obj->can('TT_LAST')
# obj.LAST(args) => $obj if ! $obj->can('TT_LAST')
# ???.LAST => ???
#------------------------------------------------------------------------

sub vmeth_last {
    my ($self, $root, $args, $nodes) = @_;
    $args ||= [];
    $self->DEBUG("LAST(root=$root, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.LAST returns single item intact
    return $root unless ref $root;

    return $root->TT_LAST(@$args)
	if UNIVERSAL::can($root, 'TT_LAST');

    $root = $self->vmeth_list($root)
	unless ref $root eq 'ARRAY';

    return $root->[-1];
}


#------------------------------------------------------------------------
# vmeth_join($root, \@args, \@nodes)	                     JOIN
#
# item.JOIN => $item
# list.JOIN(joint) => join($joint, @$list)
# hash.JOIN(joint, equals) => "$key = $value, $key = $value, ... "
# obj.JOIN(args) => $obj->TT_JOIN(@$args) if $obj->can('TT_JOIN')
# obj.JOIN(args) => $obj if ! $obj->can('TT_JOIN')
# ???.JOIN => ???
#------------------------------------------------------------------------

sub vmeth_join {
    my ($self, $root, $args, $nodes) = @_;

    # item.JOIN returns single item intact
    return $root unless ref $root;

    $args ||= [];

    return $root->TT_JOIN(@$args)
	if UNIVERSAL::can($root, 'TT_JOIN');

    my $opts = ref $args->[-1] eq 'HASH' ? pop(@$args) : { };

    # see if joint and/or equals are specified as named parameters
    # or positional arguments, otherwise these get defaulted later
    # on depending on the type of the $root item being joined
    my ($joint, $equals);
    $opts->{ joint } ||= $opts->{ join };   # might forget 't'
    $joint  = defined $opts->{ joint }  ? $opts->{ joint }  : shift @$args;
    $equals = defined $opts->{ equals } ? $opts->{ equals } : shift @$args;

    # we may need to delegate to vmeth_sort() if it looks like any
    # sort options have been specified;  in that case we also need
    # to push the named parameters options back onto the $args list
    my ($sorted, @output);
    my $do_sort = $opts->{ order } || $opts->{ item };
    push(@$args, $opts) if $do_sort;

    $self->DEBUG("JOIN(root=$root, args=[@$args], nodes...)\n")
	if $DEBUG;

    if (ref $root eq 'HASH') {
	$joint  = $HASH_JOINT unless defined $joint;
	$equals = $HASH_EQUALS unless defined $equals;

	if ($do_sort) {
	    $sorted = $self->vmeth_sort($root, $args);
	    $self->DEBUG("join sorted items: [@$sorted]\n")
		if $DEBUG;
	    while (@$sorted) {
		push(@output, sprintf("%s$equals%s", splice(@$sorted, 0, 2)));
	    }
	    $self->DEBUG("join sorted output: [@output]\n")
		if $DEBUG;
	}
	else {
	    @output = map { $_ . $equals . $root->{$_} } keys %$root;
	}

	return join($joint, @output);
    }
    elsif (ref $root ne 'ARRAY') {
	my $prev = $root;
	$root = $self->vmeth_list($root);
	$self->DEBUG("cast $prev to $root\n")
	    if $DEBUG;
    }

    $joint = $LIST_JOINT unless defined $joint;
    $root = $self->vmeth_sort($root, $args)
	if $do_sort;
    return join($joint, @$root);
}


#------------------------------------------------------------------------
# vmeth_sort($root, \@args, \@nodes)	                     JOIN
#
# item.SORT => $item
# list.SORT(options) => sort { ...options... } @$list
# hash.SORT(options) => sort { ...options... } %$hash
# obj.SORT(args) => $obj->TT_SORT(@$args) if $obj->can('TT_SORT')
# obj.SORT(args) => $obj if ! $obj->can('TT_SORT')
# ???.SORT => ???
#------------------------------------------------------------------------

sub vmeth_sort {
    my ($self, $root, $args, $nodes) = @_;
    $args ||= [];
    my $opts = ref $args->[-1] eq 'HASH' ? $args->[-1] : { };
    my ($order, $item, @result);

    $self->DEBUG("SORT(root=$root, args=[@$args], nodes...)\n")
	if $DEBUG;

    # item.SORT returns single item intact
    return $root unless ref $root;

    if (UNIVERSAL::can($root, 'TT_SORT')) {
	return $root->TT_SORT(@$args);
    }
    elsif (ref $root eq 'HASH') {
	$order = $opts->{ order } || 'alpha';
	$item = $opts->{ item } || 'keys';
	
	$self->DEBUG("sort HASH $item [", %$root, "] ordered by $order\n")
	    if $DEBUG;

	if ($item =~ /^keys?$/) {
	    if ($order eq 'alpha') {
		@result = (
		    map  { @$_[0,1] }
		    sort { $a->[2] cmp $b->[2] } 
		    map  { [ $_, $root->{$_}, lc $_ ] }
		    keys %$root 
		);
	    }
	    elsif ($order eq 'number') {
		# avoid warnings for possible non-numerical comparison items
		no warnings;
		@result = (
		    map  { @$_ }
 		    sort { $a->[0] <=> $b->[0] }
		    map  { [ $_, $root->{$_} ] }
		    keys %$root
		);
	    }
	    else {
		$self->throw("invalid sort order: $order");
	    }
	}
	elsif ($item =~  /^values?$/) {
	    if ($order eq 'alpha') {
		@result = 
		    map  { @$_[0,1] }
		    sort { $a->[2] cmp $b->[2] }
		    map  { [ $_, $root->{$_}, lc $root->{$_} ] } 
		    keys %$root;
	    }
	    elsif ($order eq 'number') {
		# avoid warnings for possible non-numerical comparison items
		no warnings;
		@result = 
		    map  { @$_ }
		    sort { $a->[1] <=> $b->[1] }
		    map  { [ $_, $root->{$_} ] } 
		    keys %$root;
	    }
	    else {
		$self->throw("invalid sort order: $order");
	    }
	}
	else {
	    $self->throw("invalid hash sort item: $item");
	}
    }
    elsif (ref $root eq 'ARRAY') {
	$order = $opts->{ order } || 'alpha';
	my $hashkey = $opts->{ hashkey } || '';

	$self->DEBUG("sort ARRAY [@$root] ordered by $order\n")
	    if $DEBUG;

	if ($order eq 'alpha') {
	    @result = (
		map  { $_->[0] }
		sort { $a->[1] cmp $b->[1] }
		map  { [ $_, lc $_ ] } 
		@$root );
	}
	elsif ($order eq 'number') {
	    no warnings;
	    @result = (
		map  { $_->[0] }
		sort { $a->[1] <=> $b->[1] }
		map  { [ $_, 0 + $_ ] } 
		@$root );
	}
	else {
	    $self->throw("invalid sort order: $order");
	}
    }
    else {
	return $root;
    }

    $self->DEBUG("sort result: [@result]\n")
	if $DEBUG;

    return \@result;
}


#------------------------------------------------------------------------
# vmeth_reverse($root, \@args, \@nodes)	                          REVERSE
#
# item.JOIN => $item
# list.JOIN(joint) => join($joint, @$list)
# hash.JOIN(joint, equals) => "$key = $value, $key = $value, ... "
# obj.JOIN(args) => $obj->TT_JOIN(@$args) if $obj->can('TT_JOIN')
# obj.JOIN(args) => $obj if ! $obj->can('TT_JOIN')
# ???.JOIN => ???
#------------------------------------------------------------------------

sub vmeth_reverse {
    my ($self, $root, $args, $nodes) = @_;

    $self->DEBUG("REVERSE(root=$root, args=[$args], nodes...)\n")
	if $DEBUG;

    # item.REVERSE returns single item intact
    return $root unless ref $root;

    if (UNIVERSAL::can($root, 'TT_REVERSE')) {
	return $root->TT_REVERSE(@$args);
    }
    elsif (ref $root ne 'ARRAY') {
	$root = $self->vmeth_list($root);
    }

    return [ reverse @$root ];
}


#------------------------------------------------------------------------
# vmeth_repeat($root, \@args, \@nodes)	                        REPEAT(n)
#
# item.REPEAT(n) => $item x $n
# list.REPEAT(n) => [ @$item x $n ] 
# hash.REPEAT(n) => hash.LIST.REPEAT(n)
# obj.REPEAT(args) => $obj->TT_REPEAT(@$args) if $obj->can('TT_REPEAT')
# obj.REPEAT(args) => $obj if ! $obj->can('TT_REPEAT')
# ???.REPEAT(n) => ???.LIST.REPEAT
#------------------------------------------------------------------------

sub vmeth_repeat {
    my ($self, $root, $args, $nodes) = @_;
    $args ||= [ ];
    my $n = shift @$args;
    $n = 1 unless defined $n;

    $self->DEBUG("REPEAT(root=$root, n=$n, nodes...)\n")
	if $DEBUG;

    return $root unless $n > 1;
    return $root x $n unless ref $root;

    if (UNIVERSAL::can($root, 'TT_REPEAT')) {
	return $root->TT_REPEAT(@$args);
    }
    elsif (ref $root ne 'ARRAY') {
	$root = $self->vmeth_list($root);
    }

    return [ (@$root) x $n ];
}


#========================================================================

#------------------------------------------------------------------------
# update(\%params)
#
# Update multiple variables en masse.  No magic is performed.  Simple
# variable names only.
#------------------------------------------------------------------------

sub update {
    my ($self, $params) = @_;

    # look out for magical 'import' argument to import another hash
    my $import = $params->{ import };
    if (defined $import && UNIVERSAL::isa($import, 'HASH')) {
	@$self{ keys %$import } = values %$import;
	delete $params->{ import };
    }

    @$self{ keys %$params } = values %$params;
}



sub WARNING {
    my $self = shift;
    print STDERR join('', @_)
	if $WARNINGS;
}


#------------------------------------------------------------------------
# _dump()
#
# Debug method which returns a string representing the internal state
# of the object.  The method calls itself recursively to dump sub-hashes.
#------------------------------------------------------------------------

sub _dump {
    my $self   = shift;
    my $indent = shift || 1;
    my $buffer = '    ';
    my $pad    = $buffer x $indent;
    my $text   = '';
    local $" = ', ';

    my ($key, $value);


    return $text . "...excessive recursion, terminating\n"
	if $indent > 32;

    foreach $key (keys %$self) {

	$value = $self->{ $key };
	$value = '<undef>' unless defined $value;

	if (ref($value) eq 'ARRAY') {
	    $value = "$value [@$value]";
	}
	$text .= sprintf("$pad%-8s => $value\n", $key);
	next if $key =~ /^\./;
	if (UNIVERSAL::isa($value, 'HASH')) {
	    $text .= _dump($value, $indent + 1);
	}
    }
    $text;
}

sub DEBUG {
    my $self = shift;
    print STDERR @_;
}


1;

__END__

