#!/usr/bin/perl -w                                         # -*- perl -*-
#
# t/parse01.t
#
# Test parser's ability to parse miscellaneous elements of the TT language:
# filenames, etc.
#

use strict;
use lib qw( ../lib ./lib );
use Template::Test;
use Template::Parser;

use vars qw( $DEBUG );
$DEBUG = 0;
$Template::Parser::DEBUG = $DEBUG;
local $" = ', ';
local $/ = undef;

#------------------------------------------------------------------------
# Define our own debugging backend so that we have control over what the 
# parser generates.

package My::Compiler;
use Template::Compiler;
use base qw( Template::Compiler );

sub number {
    my ($self, $number) = @_;
    return "number<$number>";
}

sub literal {
    my ($self, $literal) = @_;
    return "literal<'$literal'>";
}

sub quoted {
    my ($self, $text) = @_;
    return "quoted<@$text>";
}

sub variable {
    my ($self, $vars) = @_;
    local $" = ', ';
    return "variable<@$vars>";
}

sub assign {
    my ($self, $var, $value) = @_;
    local $" = ', ';
    return "assign<[@$var] => $value>";
}

sub list {
    my ($self, $list) = @_;
    local $" = ', ';
    return "[ @$list ]";
}

sub args {
    my ($self, $args) = @_;
    my $hash = shift @$args || [];
    push(@$args, $self->hash($hash))
	if @$hash;
    return $self->list($args);
}

sub hash {
    my ($self, $hash) = @_;
    local $" = ', ';
    return "{ @$hash }";
}

sub filename {
    my ($self, $text) = @_;
    return "filename<$text>";
}


#------------------------------------------------------------------------
package main;
#------------------------------------------------------------------------

# read tests, strip comments and split into list of [ source => expect ]
my $tests = <DATA>;
$tests =~ s/.*?__start__\n//s;
$tests = [ map { s/^(#.*\n)+//g; [ split /\n/ ] } split(/\n\n/, $tests) ];

# create a parser installed with our debug backend
my $parser = Template::Parser->new( COMPILER => My::Compiler->new() );
assert( $parser );

my ($test, $expect);

my $source = 'foo';
my $result = $parser->parse_term($source);

match( $result, "variable<literal<'foo'>, 0>" );

$result = $parser->parse_term(\$source);
match( $result, "variable<literal<'foo'>, 0>" );

$result = $parser->parse_term(\$source);
ok( ! $result );

# reset regex position to start
pos $source = 0;

$result = $parser->parse_term(\$source);
match( $result, "variable<literal<'foo'>, 0>" );

$result = $parser->parse_term('4');
match( $result, "number<4>" );

$result = $parser->parse_term('$foo');
match( $result, "variable<variable<literal<'foo'>, 0>, 0>" );

# TODO
#$result = $parser->parse_term('${foo}');
#match( $result, "var<embed<foo>, 0>" );
#
#$result = $parser->parse_term('${foo.bar}');
#match( $result, "var<embed<foo.bar>, 0>" );

#------------------------------------------------------------------------


# process each test, parsing source as an expression and testing 
# generated output against expected.

foreach $test (@$tests) {
    ($source, $expect) = @$test;
    print STDERR "source [$source]  expect [$expect]\n"
	if $DEBUG;
    $result = $parser->parse_filename($source);
    unless (defined $result) {
	ok(0, "Parser returned an undefined value\n" . $parser->error());
	next;
    }
    match( $result, $expect );
}

__END__
foo
literal<'foo'>

foo/bar
literal<'foo/bar'>

foo/bar.baz
literal<'foo/bar.baz'>

/foo/bar/baz.txt
literal<'/foo/bar/baz.txt'>

../baz.txt
literal<'../baz.txt'>

Foo::Bar
literal<'Foo::Bar'>

Foo::Bar.pm
literal<'Foo::Bar.pm'>

"foo"
quoted<"foo">

"foo/bar.pm"
quoted<"foo/bar.pm">

'foo/bar.pm'
literal<'foo/bar.pm'>

