#!/usr/bin/perl -w                                         # -*- perl -*-
#
# t/parser00.t
#
# Test parser's ability to parse various elements of the TT language,
# e.g. expressions, variables, quoted text, lists, hash arrays, etc.
#

use strict;
use lib qw( ../lib ./lib );
use Template::Test;
use Template::Parser;

use vars qw( $DEBUG );
$DEBUG = 0;
$Template::Parser::DEBUG = $DEBUG;
local $" = ', ';
local $/ = undef;


#------------------------------------------------------------------------
# Define our own debugging backend so that we have control over what the 
# parser generates.

package My::Backend;
use Template::Compiler;
use base qw( Template::Compiler );

sub number {
    my ($self, $number) = @_;
    return "number<$number>";
}

sub literal {
    my ($self, $literal) = @_;
    return "literal<'$literal'>";
}

sub quoted {
    my ($self, $text) = @_;
    return "quoted<@$text>";
}

sub variable {
    my ($self, $vars) = @_;
    local $" = ', ';
    return "variable<@$vars>";
}

sub assign {
    my ($self, $var, $value) = @_;
    local $" = ', ';
    return "assign<[@$var] => $value>";
}

sub list {
    my ($self, $list) = @_;
    local $" = ', ';
    return "[ @$list ]";
}

sub args {
    my ($self, $args) = @_;
    my $hash = shift @$args || [];
    push(@$args, $self->hash($hash))
	if @$hash;
    return $self->list($args);
}

sub hash {
    my ($self, $hash) = @_;
    local $" = ', ';
    return "{ @$hash }";
}





#------------------------------------------------------------------------
package main;

# read tests, strip comments and split into list of [ source => expect ]
my $tests = <DATA>;
$tests =~ s/.*?__start__\n//s;
$tests = [ map { s/^(#.*\n)+//g; [ split /\n/ ] } split(/\n\n/, $tests) ];

# create a parser installed with our debug backend
my $parser = Template::Parser->new( COMPILER => My::Backend->new());
assert( $parser );

my $compiler = $parser->{ _COMPILER };
my $handler = $compiler->template_handler();
$parser->push_handler($handler);

my ($test, $source, $result, $expect);

# process each test, parsing source as an expression and testing 
# generated output against expected.

foreach $test (@$tests) {
    ($source, $expect) = @$test;
    print STDERR "source [$source]  expect [$expect]\n"
	if $DEBUG;
    $result = $parser->parse_expression($source);
    unless (defined $result) {
	ok(0, "Parser returned an undefined value\n");
	next;
    }
    match( $result, $expect );
}

__END__
#------------------------------------------------------------------------
# test numbers
-1
number<-1>

0
number<0>

1
number<1>

+2
number<+2>

2.718
number<2.718>

-2.718
number<-2.718>

+3.14159
number<+3.14159>

#------------------------------------------------------------------------
# test literal strings
'foo'
literal<'foo'>

'foo$bar'
literal<'foo$bar'>

#------------------------------------------------------------------------
# test double quoted strings    TODO: fix this when working
"foo"
quoted<"foo">

"foo $bar \"hello\"\\"
quoted<"foo ", variable<literal<'bar'>, 0>, " \"hello\"\\">

#------------------------------------------------------------------------
# test variables
foo
variable<literal<'foo'>, 0>

bar(a)
variable<literal<'bar'>, [ variable<literal<'a'>, 0> ]>

bar(a, b)
variable<literal<'bar'>, [ variable<literal<'a'>, 0>, variable<literal<'b'>, 0> ]>

bar(a b)
variable<literal<'bar'>, [ variable<literal<'a'>, 0>, variable<literal<'b'>, 0> ]>

foo.bar
variable<literal<'foo'>, 0, literal<'bar'>, 0>

foo('a').bar(b)
variable<literal<'foo'>, [ literal<'a'> ], literal<'bar'>, [ variable<literal<'b'>, 0> ]>

foo.3
variable<literal<'foo'>, 0, number<3>, 0>

foo.$bar
variable<literal<'foo'>, 0, variable<literal<'bar'>, 0>, 0>

foo.${bar}
variable<literal<'foo'>, 0, do { variable<literal<'bar'>, 0> }, 0>

foo.${bar.baz}
variable<literal<'foo'>, 0, do { variable<literal<'bar'>, 0, literal<'baz'>, 0> }, 0>

foo.${'bar'}
variable<literal<'foo'>, 0, do { literal<'bar'> }, 0>

#------------------------------------------------------------------------
# test list
[ foo ]
[ variable<literal<'foo'>, 0> ]

[ 1, 2, 3 ]
[ number<1>, number<2>, number<3> ]

[ 1 2 3 ]
[ number<1>, number<2>, number<3> ]

[ 1 [ 2 [ 3 ] ] ]
[ number<1>, [ number<2>, [ number<3> ] ] ]

#------------------------------------------------------------------------
# test hash
{ foo = bar 'baz' => 123, a=b }
{ literal<'foo'>, variable<literal<'bar'>, 0>, literal<'baz'>, number<123>, literal<'a'>, variable<literal<'b'>, 0> }

#------------------------------------------------------------------------
# test expressions
1 ne 2 eq 3
number<1> ne number<2> eq number<3>

#------------------------------------------------------------------------
# test expressions
1 && 2 || ! 3 && 4
number<1> && number<2> || ! number<3> && number<4>

foo && bar ? baz > 3 ? 1 : 2 : 3
variable<literal<'foo'>, 0> && variable<literal<'bar'>, 0> ? variable<literal<'baz'>, 0> > number<3> ? number<1> : number<2> : number<3>

! foo || ! bar 
! variable<literal<'foo'>, 0> || ! variable<literal<'bar'>, 0>

(a && b) || (c)
(variable<literal<'a'>, 0> && variable<literal<'b'>, 0>) || (variable<literal<'c'>, 0>)

(foo = bar)
assign<[literal<'foo'>, 0] => variable<literal<'bar'>, 0>>

(foo = bar) ? a : b
assign<[literal<'foo'>, 0] => variable<literal<'bar'>, 0>> ? variable<literal<'a'>, 0> : variable<literal<'b'>, 0>

(foo.bar = [ 1 2 3 ])
assign<[literal<'foo'>, 0, literal<'bar'>, 0] => [ number<1>, number<2>, number<3> ]>

#------------------------------------------------------------------------
# test args
foo(a, b)
variable<literal<'foo'>, [ variable<literal<'a'>, 0>, variable<literal<'b'>, 0> ]>

foo(a => b)
variable<literal<'foo'>, [ { literal<'a'>, variable<literal<'b'>, 0> } ]>

foo(a, b => c)
variable<literal<'foo'>, [ variable<literal<'a'>, 0>, { literal<'b'>, variable<literal<'c'>, 0> } ]>

foo(a b = c)
variable<literal<'foo'>, [ variable<literal<'a'>, 0>, { literal<'b'>, variable<literal<'c'>, 0> } ]>

foo(a b = c,)
variable<literal<'foo'>, [ variable<literal<'a'>, 0>, { literal<'b'>, variable<literal<'c'>, 0> } ]>

file://tmp/foo.txt
file_resource(//tmp/foo.txt)

http://www.foo.com/bar.xml
http_resource(//www.foo.com/bar.xml)
