#============================================================= -*-perl-*-
#
# t/stash01.t
#
# Test the Template::Stash::v3 module from templates.
#
# Written by Andy Wardley <abw@kfs.org>
#
# Copyright (C) 1996-2001 Andy Wardley.  All Rights Reserved.
# Copyright (C) 1998-2001 Canon Research Centre Europe Ltd.
#
# This is free software; you can redistribute it and/or modify it
# under the same terms as Perl itself.
#
# $Id: stash01.t,v 1.1 2001/06/27 13:37:48 abw Exp $
#
#========================================================================

use strict;

# NOTE: this tests v3 stash in a v2 framework so make sure that it finds
# V2 file *before* any of the v3 files.
use lib qw( /home/abw/tt2/lib ../lib ./lib );

use Template;
use Template::Test;
use Template::Stash::v3;

local $/ = undef;
my $DEBUG  = 0;
my $ITER   = 1000;
my $source = 'test/stash';

#------------------------------------------------------------------------
package Foo;
#------------------------------------------------------------------------

sub new {
    my $class = shift;
    bless { foo => 'bar', wiz => 'waz' }, $class;
}

sub magic {
    my $self = shift;
    my @args = @_;
    if (wantarray) {
	return map { "more_magic($_)" } @args;
    }
    else {
	return "less: " . join(', ', map { "magic($_)" } @args);
    }
}


#------------------------------------------------------------------------
package main;
#------------------------------------------------------------------------

my @days = qw( Monday Tuesday Wednesday Thursday Friday Saturday Sunday );
my ($a, $b, $c, $d, $e, $f, $g, $h, $i, $j, $k, $l, $m, 
    $n, $o, $p, $q, $r, $s, $t, $u, $v, $w, $x, $y, $z) = 
	qw( alpha bravo charlie delta echo foxtrot golf hotel india 
	    juliet kilo lima mike november oscar papa quebec romeo 
	    sierra tango umbrella victor whisky x-ray yankee zulu );

my $vars = { 
    a => $a,
    b => {
	c => $c,
	d => {
	    e => $e,
	    f => sub { return "foo(" . join(', ', @_) . ")" },
	    g => [ qw( zero one two three ) ],
	},
	hw => {
	    h => 'hello',
	    w => 'world',
	},  
    },
    days  => \@days,
    magic => \&magic,
    more  => Foo->new(),
    debug => sub { $Template::Stash::v3::DEBUG = shift },
    warn  => sub { $Template::Stash::v3::WARNINGS = shift },
};

sub magic {
    if (wantarray) {
	return qw( hocus pocus );
    }
    else {
	return 'Abracadabra';
    }
}

my $stash_v3 = Template::Stash::v3->new($vars);
my $tt_v3    = Template->new( STASH => $stash_v3 );

test_expect(\*DATA, $tt_v3);

__END__
-- test --
[% a %]
-- expect --
alpha

#------------------------------------------------------------------------
# testing calling context (ARRAY vs SCALAR) via use of trailing LIST/ITEM
# vmeth
#------------------------------------------------------------------------
-- test --
[% magic.JOIN() %]
-- expect --
hocus, pocus

-- test --
[% magic.JOIN(', ') %]
-- expect --
hocus, pocus

# default to list context
-- test --
[% magic.JOIN(' + ') %]
-- expect --
hocus + pocus

#------------------------------------------------------------------------
# brief diversion to play with JOIN... (which I just wrote to get the last 
# tests to work... :-)
#------------------------------------------------------------------------
-- test --
[% days.JOIN %]
-- expect --
Monday, Tuesday, Wednesday, Thursday, Friday, Saturday, Sunday

-- test --
'[% days.JOIN("', '") %]'
-- expect --
'Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday'

-- test --
[% b.hw.JOIN( order='alpha' ) %]
-- expect --
h = hello, w = world

-- test --
[% b.hw.JOIN(",\n", " => ", order='alpha') %]
-- expect --
h => hello,
w => world

#------------------------------------------------------------------------
# OK, back to calling context... now try explicit LIST context
#------------------------------------------------------------------------
-- test --
[% magic.LIST.JOIN(' + ') %]
-- expect --
hocus + pocus

# ITEM implies scalar context 
-- test --
[% magic.ITEM %]
-- expect --
Abracadabra

# then let JOIN cast it up to a list of 1 item 
-- test --
[% magic.ITEM.JOIN(' + ') %]
-- expect --
Abracadabra

# now try that object (disable warning because undef value is returned)
-- test --
[% CALL warn(0); more.magic.JOIN; CALL warn(1) %].
-- expect --
.

-- test --
[% more.magic(1).JOIN %]
-- expect --
more_magic(1)

-- test --
[% more.magic(3.14, 2.718).JOIN %]
-- expect --
more_magic(3.14), more_magic(2.718)

-- test --
[% more.magic(3.14, 2.718).ITEM %]
-- expect --
less: magic(3.14), magic(2.718)

#------------------------------------------------------------------------
# REF tells us what kind of data an item is
#------------------------------------------------------------------------
-- test --
[% a.REF %]?
-- expect --
?

-- test --
[% b.c.REF %]?
-- expect --
?

-- test --
[% b.d.REF %]
-- expect --
HASH

-- test --
[% b.hw.REF %]
-- expect --
HASH

-- test --
[% magic.REF %]
-- expect --
CODE

-- test --
[% more.REF %]
-- expect --
Foo

-- test --
[% more.magic.REF %]
-- expect --
CODE(Foo)



#------------------------------------------------------------------------
# CODEREF wraps up a variable in a closure for lazy evaluation
#------------------------------------------------------------------------

-- test --
[% runme = a.CODEREF;
   runme %]
-- expect --
alpha

-- test --
[% runme = b.c.CODEREF;
   runme %]
-- expect --
charlie

-- test --
[% # you can create a CODEREF to wrap an object method.
   runme = more.magic.CODEREF;
#  CALL debug(1);
   runme('wamblam')
%]
-- expect --
more_magic(wamblam)

-- test --
[% # you can also trap arguments in the closure
   runme = more.magic('Hello World!').CODEREF;
#  CALL debug(1);
   runme('wamblam').JOIN
%]
-- expect --
more_magic(Hello World!), more_magic(wamblam)


